import pytest
from pytest import param
from freezegun import freeze_time
from bumpversion.versioning.functions import NumericFunction, ValuesFunction, IndependentFunction, CalVerFunction


# NumericFunction
class TestNumericFunction:
    """The numeric function manages incrementing and resetting numeric version parts."""

    class TestCreation:
        """The numeric function can be created."""

        def test_without_first_value_is_zero(self):
            func = NumericFunction()
            assert func.first_value == "0"

        def test_with_first_value_is_value(self):
            func = NumericFunction(first_value="5")
            assert func.first_value == "5"

        def test_non_numeric_first_value_raises_value_error(self):
            with pytest.raises(ValueError):
                NumericFunction(first_value="a")

        def test_alphanumeric_value_is_accepted(self):
            func = NumericFunction(first_value="v10b")
            assert func.first_value == "v10b"

        def test_without_optional_value_is_first_value(self):
            func = NumericFunction()
            assert func.optional_value == func.first_value == "0"
            func = NumericFunction(first_value="5")
            assert func.optional_value == func.first_value == "5"

    class TestBump:
        """Tests for incrementing a value."""

        def test_number_increments(self):
            func = NumericFunction()
            assert func.bump("0") == "1"

        @pytest.mark.parametrize(
            ("value", "expected"),
            [
                param("v10", "v11", id="alphanumeric-prefix"),
                param("10b", "11b", id="alphanumeric-suffix"),
                param("v10b", "v11b", id="alphanumeric-prefix-suffix"),
                param("r3-001", "r4-001", id="alphanumeric-prefix-suffix-with-numeric-suffix"),
            ],
        )
        def test_alphanumeric_increments(self, value: str, expected: str):
            func = NumericFunction()
            assert func.bump(value) == expected

        def test_bump_before_first_value_raises_value_error(self):
            func = NumericFunction()
            with pytest.raises(ValueError):
                func.bump("-1")

            func = NumericFunction(first_value="5")
            with pytest.raises(ValueError):
                func.bump("3")

        def test_bump_non_numeric_value_raises_value_error(self):
            func = NumericFunction()
            with pytest.raises(ValueError):
                func.bump("a")


# ValuesFunction
class TestValuesFunction:
    """The values function manages incrementing and resetting non-numeric version parts."""

    class TestCreation:
        """The values function can be created."""

        def test_values_set_first_value_and_optional_value(self):
            func = ValuesFunction(["0", "1", "2"])
            assert func.optional_value == "0"
            assert func.first_value == "0"

        def test_optional_value_can_be_different_from_first_value(self):
            func = ValuesFunction(["0", "1", "2"], optional_value="1")
            assert func.optional_value == "1"
            assert func.first_value == "0"

            func = ValuesFunction(["0", "1", "2"], optional_value="0", first_value="1")
            assert func.optional_value == "0"
            assert func.first_value == "1"

        def test_optional_value_must_be_in_values(self):
            with pytest.raises(ValueError):
                ValuesFunction(["0", "1", "2"], optional_value="3")

        def test_first_value_doesnt_need_to_be_first_in_values(self):
            func = ValuesFunction(["0", "1", "2"], first_value="1")
            assert func.optional_value == "0"
            assert func.first_value == "1"

        def test_values_are_required(self):
            with pytest.raises(ValueError):
                ValuesFunction([])

        def test_first_value_must_be_in_values(self):
            with pytest.raises(ValueError):
                ValuesFunction(["0", "1", "2"], first_value="3")

    class TestBump:
        def test_bump_returns_next_item(self):
            func = ValuesFunction(["0", "5", "10"])
            assert func.bump("0") == "5"

        def test_cannot_bump_beyond_max_value(self):
            func = ValuesFunction(["0", "5", "10"])
            with pytest.raises(ValueError):
                func.bump("10")

        def test_bump_non_value_raises_value_error(self):
            func = ValuesFunction(["0", "5", "10"])
            with pytest.raises(ValueError):
                func.bump("3")


class TestIndependentFunction:
    """The independent function manages incrementing and resetting version parts."""

    class TestCreation:
        """The independent function can be created."""

        def test_value_sets_first_value_and_optional_value(self):
            func = IndependentFunction("value")
            assert func.optional_value == "value"
            assert func.first_value == "value"

        def test_value_is_not_required(self):
            assert IndependentFunction().optional_value == ""
            assert IndependentFunction("").optional_value == ""

    class TestBump:
        def test_bump_with_value_returns_value(self):
            func = IndependentFunction("1")
            assert func.bump("5") == "5"

        def test_bump_with_no_value_returns_initial_value(self):
            func = IndependentFunction("1")
            assert func.bump() == "1"


class TestCalVerFunction:
    """The calver function manages incrementing and resetting calver version parts."""

    @freeze_time("2020-05-01")
    def test_creation_sets_first_value_and_optional_value(self):
        func = CalVerFunction("{YYYY}.{MM}")
        assert func.optional_value == "There isn't an optional value for CalVer."
        assert func.first_value == "2020.5"
        assert func.calver_format == "{YYYY}.{MM}"

    @freeze_time("2020-05-01")
    def test_bump_with_value_ignores_value(self):
        func = CalVerFunction("{YYYY}.{MM}.{DD}")
        assert func.bump("123456") == "2020.5.1"

    @pytest.mark.parametrize(
        ["calver", "expected"],
        [
            param("{YYYY}", "2002", id="{YYYY}"),
            param("{YY}", "2", id="{YY}"),
            param("{0Y}", "02", id="{0Y}"),
            param("{MMM}", "May", id="{MMM}"),
            param("{MM}", "5", id="{MM}"),
            param("{0M}", "05", id="{0M}"),
            param("{DD}", "1", id="{DD}"),
            param("{0D}", "01", id="{0D}"),
            param("{JJJ}", "121", id="{JJJ}"),
            param("{00J}", "121", id="{00J}"),
            param("{Q}", "2", id="{Q}"),
            param("{WW}", "17", id="{WW}"),
            param("{0W}", "17", id="{0W}"),
            param("{UU}", "17", id="{UU}"),
            param("{0U}", "17", id="{0U}"),
            param("{VV}", "18", id="{VV}"),
            param("{0V}", "18", id="{0V}"),
            param("{GGGG}", "2002", id="{GGGG}"),
            param("{GG}", "2", id="{GG}"),
            param("{0G}", "02", id="{0G}"),
        ],
    )
    @freeze_time("2002-05-01")
    def test_calver_formatting_renders_correctly(self, calver: str, expected: str):
        """Test that the calver is formatted correctly."""
        func = CalVerFunction(calver)
        assert func.bump() == expected

    @pytest.mark.parametrize(
        ["calver", "expected"],
        [
            param("{YYYY}", "2000", id="{YYYY}"),
            param("{YY}", "0", id="{YY}"),
            param("{0Y}", "00", id="{0Y}"),
            param("{GGGG}", "1999", id="{GGGG}"),
            param("{GG}", "99", id="{GG}"),
            param("{0G}", "99", id="{0G}"),
        ],
    )
    @freeze_time("2000-01-01")
    def test_century_years_return_zeros(self, calver: str, expected: str):
        func = CalVerFunction(calver)
        assert func.bump() == expected



"""Visualize the bumpversion process."""

from dataclasses import dataclass
from typing import List, Optional

from bumpversion.bump import get_next_version
from bumpversion.config import Config
from bumpversion.context import base_context, get_context
from bumpversion.exceptions import BumpVersionError
from bumpversion.ui import print_info

BOX_CHARS = {
    "ascii": ["+", "+", "+", "+", "+", "+", "+", "+", "-", "|", "+"],
    "light": ["╯", "╮", "╭", "╰", "┤", "┴", "┬", "├", "─", "│", "┼"],
}


@dataclass
class Border:
    """A border definition."""

    corner_bottom_right: str
    corner_top_right: str
    corner_top_left: str
    corner_bottom_left: str
    divider_left: str
    divider_up: str
    divider_down: str
    divider_right: str
    line: str
    pipe: str
    cross: str


def lead_string(version_str: str, border: Border, blank: bool = False) -> str:
    """
    Return the first part of a string with the bump character or spaces of the correct amount.

    Examples:
        >>> lead_string("1.0.0", Border(*BOX_CHARS["light"]))
        '1.0.0 ── bump ─'
        >>> lead_string("1.0.0", Border(*BOX_CHARS["light"]), blank=True)
        '               '

    Args:
        version_str: The string to render as the starting point
        border: The border definition to draw the lines
        blank: If `True`, return a blank string the same length as the version bump string

    Returns:
        The version bump string or a blank string
    """
    version_bump = f"{version_str} {border.line * 2} bump {border.line}"
    return " " * len(version_bump) if blank else version_bump


def connection_str(border: Border, has_next: bool = False, has_previous: bool = False) -> str:
    """
    Return the correct connection string based on the next and previous.

    Args:
        border: The border definition to draw the lines
        has_next: If `True`, there is a next line
        has_previous: If `True`, there is a previous line

    Returns:
        A string that connects left-to-right and top-to-bottom based on the next and previous
    """
    if has_next and has_previous:
        return border.divider_right + border.line
    elif has_next:
        return border.divider_down + border.line
    elif has_previous:
        return border.corner_bottom_left + border.line
    else:
        return border.line * 2


def labeled_line(label: str, border: Border, fit_length: Optional[int] = None) -> str:
    """
    Return the version part string with the correct padding.

    Args:
        label: The label to render
        border: The border definition to draw the lines
        fit_length: The length to fit the label to

    Returns:
        A labeled line with leading and trailing spaces
    """
    if fit_length is None:
        fit_length = len(label)
    return f" {label} {border.line * (fit_length - len(label))}{border.line} "


def filter_version_parts(config: Config) -> List[str]:
    """
    Return the version parts that are in the configuration.

    Args:
        config: The configuration to check against

    Returns:
        The version parts that are in the configuration
    """
    version_parts = [part for part in config.version_config.order if not part.startswith("$")]
    default_context = base_context(config.scm_info)
    return [part for part in version_parts if part not in default_context]


def visualize(config: Config, version_str: str, box_style: str = "light") -> None:
    """Output a visualization of the bump-my-version bump process."""
    version = config.version_config.parse(version_str)
    version_parts = filter_version_parts(config)
    num_parts = len(version_parts)

    box_style = box_style if box_style in BOX_CHARS else "light"
    border = Border(*BOX_CHARS[box_style])

    version_lead = lead_string(version_str, border)
    blank_lead = lead_string(version_str, border, blank=True)
    version_part_length = max(len(part) for part in version_parts)

    for i, part in enumerate(version_parts):
        line = [version_lead] if i == 0 else [blank_lead]

        try:
            next_version = get_next_version(version, config, part, None)
            next_version_str = config.version_config.serialize(next_version, get_context(config))
        except (BumpVersionError, ValueError) as e:
            next_version_str = f"invalid: {e}"

        has_next = i < num_parts - 1
        has_previous = i > 0
        line.append(connection_str(border, has_next=has_next, has_previous=has_previous))
        line.append(labeled_line(part, border, version_part_length))
        line.append(next_version_str)
        print_info("".join(line))



"""Version control system management."""

import os
import re
import subprocess
from dataclasses import dataclass
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, ClassVar, List, MutableMapping, Optional, Type, Union

from bumpversion.ui import get_indented_logger
from bumpversion.utils import extract_regex_flags

if TYPE_CHECKING:  # pragma: no-coverage
    from bumpversion.config import Config

from bumpversion.exceptions import BumpVersionError, DirtyWorkingDirectoryError, SignedTagsError

logger = get_indented_logger(__name__)


@dataclass
class SCMInfo:
    """Information about the current source code manager and state."""

    tool: Optional[Type["SourceCodeManager"]] = None
    commit_sha: Optional[str] = None
    distance_to_latest_tag: int = 0
    current_version: Optional[str] = None
    branch_name: Optional[str] = None
    short_branch_name: Optional[str] = None
    dirty: Optional[bool] = None

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        tool_name = self.tool.__name__ if self.tool else "No SCM tool"
        return (
            f"SCMInfo(tool={tool_name}, commit_sha={self.commit_sha}, "
            f"distance_to_latest_tag={self.distance_to_latest_tag}, current_version={self.current_version}, "
            f"dirty={self.dirty})"
        )


class SourceCodeManager:
    """Base class for version control systems."""

    _TEST_USABLE_COMMAND: ClassVar[List[str]] = []
    _COMMIT_COMMAND: ClassVar[List[str]] = []
    _ALL_TAGS_COMMAND: ClassVar[List[str]] = []

    @classmethod
    def commit(cls, message: str, current_version: str, new_version: str, extra_args: Optional[list] = None) -> None:
        """Commit the changes."""
        extra_args = extra_args or []
        if not current_version:
            logger.warning("No current version given, using an empty string.")
            current_version = ""
        if not new_version:
            logger.warning("No new version given, using an empty string.")
            new_version = ""

        with NamedTemporaryFile("wb", delete=False) as f:
            f.write(message.encode("utf-8"))

        env = os.environ.copy()
        env["HGENCODING"] = "utf-8"
        env["BUMPVERSION_CURRENT_VERSION"] = current_version
        env["BUMPVERSION_NEW_VERSION"] = new_version

        try:
            cmd = [*cls._COMMIT_COMMAND, f.name, *extra_args]
            subprocess.run(cmd, env=env, capture_output=True, check=True)  # noqa: S603
        except (subprocess.CalledProcessError, TypeError) as exc:  # pragma: no-coverage
            if isinstance(exc, TypeError):
                err_msg = f"Failed to run {cls._COMMIT_COMMAND}: {exc}"
            else:
                err_msg = f"Failed to run {exc.cmd}: return code {exc.returncode}, output: {exc.output}"
            logger.exception(err_msg)
            raise BumpVersionError(err_msg) from exc
        finally:
            os.unlink(f.name)

    @classmethod
    def is_usable(cls) -> bool:
        """Is the VCS implementation usable."""
        try:
            result = subprocess.run(cls._TEST_USABLE_COMMAND, check=True, capture_output=True)  # noqa: S603
            return result.returncode == 0
        except (FileNotFoundError, PermissionError, NotADirectoryError, subprocess.CalledProcessError):
            return False

    @classmethod
    def assert_nondirty(cls) -> None:
        """Assert that the working directory is not dirty."""
        raise NotImplementedError()

    @classmethod
    def latest_tag_info(cls, tag_name: str, parse_pattern: str) -> SCMInfo:
        """Return information about the latest tag."""
        raise NotImplementedError()

    @classmethod
    def add_path(cls, path: Union[str, Path]) -> None:
        """Add a path to the VCS."""
        raise NotImplementedError()

    @classmethod
    def tag(cls, name: str, sign: bool = False, message: Optional[str] = None) -> None:
        """Create a tag of the new_version in VCS."""
        raise NotImplementedError

    @classmethod
    def get_all_tags(cls) -> List[str]:
        """Return all tags in VCS."""
        try:
            result = subprocess.run(cls._ALL_TAGS_COMMAND, text=True, check=True, capture_output=True)  # noqa: S603
            return result.stdout.splitlines()
        except (FileNotFoundError, PermissionError, NotADirectoryError, subprocess.CalledProcessError):
            return []

    @classmethod
    def get_version_from_tag(cls, tag: str, tag_name: str, parse_pattern: str) -> Optional[str]:
        """Return the version from a tag."""
        version_pattern = parse_pattern.replace("\\\\", "\\")
        version_pattern, regex_flags = extract_regex_flags(version_pattern)
        rep = tag_name.replace("{new_version}", f"(?P<current_version>{version_pattern})")
        rep = f"{regex_flags}{rep}"
        tag_regex = re.compile(rep)
        return match["current_version"] if (match := tag_regex.match(tag)) else None

    @classmethod
    def commit_to_scm(
        cls,
        files: List[Union[str, Path]],
        config: "Config",
        context: MutableMapping,
        extra_args: Optional[List[str]] = None,
        dry_run: bool = False,
    ) -> None:
        """Commit the files to the source code management system."""
        if not cls.is_usable():
            logger.error("SCM tool '%s' is unusable, unable to commit.", cls.__name__)
            return

        if not config.commit:
            logger.info("Would not commit")
            return

        do_commit = not dry_run
        logger.info(
            "%s %s commit",
            "Preparing" if do_commit else "Would prepare",
            cls.__name__,
        )
        logger.indent()
        for path in files:
            logger.info(
                "%s changes in file '%s' to %s",
                "Adding" if do_commit else "Would add",
                path,
                cls.__name__,
            )

            if do_commit:
                cls.add_path(path)

        commit_message = config.message.format(**context)

        logger.info(
            "%s to %s with message '%s'",
            "Committing" if do_commit else "Would commit",
            cls.__name__,
            commit_message,
        )
        if do_commit:
            cls.commit(
                message=commit_message,
                current_version=context["current_version"],
                new_version=context["new_version"],
                extra_args=extra_args,
            )
        logger.dedent()

    @classmethod
    def tag_in_scm(cls, config: "Config", context: MutableMapping, dry_run: bool = False) -> None:
        """Tag the current commit in the source code management system."""
        if not config.tag:
            logger.info("Would not tag")
            return
        sign_tags = config.sign_tags
        tag_name = config.tag_name.format(**context)
        tag_message = config.tag_message.format(**context)
        existing_tags = cls.get_all_tags()
        do_tag = not dry_run

        if tag_name in existing_tags:
            logger.warning("Tag '%s' already exists. Will not tag.", tag_name)
            return

        logger.info(
            "%s '%s' %s in %s and %s",
            "Tagging" if do_tag else "Would tag",
            tag_name,
            f"with message '{tag_message}'" if tag_message else "without message",
            cls.__name__,
            "signing" if sign_tags else "not signing",
        )
        if do_tag:
            cls.tag(tag_name, sign_tags, tag_message)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"{self.__class__.__name__}"


class Git(SourceCodeManager):
    """Git implementation."""

    _TEST_USABLE_COMMAND: ClassVar[List[str]] = ["git", "rev-parse", "--git-dir"]
    _COMMIT_COMMAND: ClassVar[List[str]] = ["git", "commit", "-F"]
    _ALL_TAGS_COMMAND: ClassVar[List[str]] = ["git", "tag", "--list"]

    @classmethod
    def assert_nondirty(cls) -> None:
        """Assert that the working directory is not dirty."""
        lines = [
            line.strip()
            for line in subprocess.check_output(["git", "status", "--porcelain"]).splitlines()  # noqa: S603, S607
            if not line.strip().startswith(b"??")
        ]

        if lines:
            joined_lines = b"\n".join(lines).decode()
            raise DirtyWorkingDirectoryError(f"Git working directory is not clean:\n\n{joined_lines}")

    @classmethod
    def latest_tag_info(cls, tag_name: str, parse_pattern: str) -> SCMInfo:
        """Return information about the latest tag."""
        try:
            # git-describe doesn't update the git-index, so we do that
            subprocess.run(["git", "update-index", "--refresh", "-q"], capture_output=True)  # noqa: S603, S607
        except subprocess.CalledProcessError as e:
            logger.debug("Error when running git update-index: %s", e.stderr)
            return SCMInfo(tool=cls)
        tag_pattern = tag_name.replace("{new_version}", "*")
        try:
            # get info about the latest tag in git
            git_cmd = [
                "git",
                "describe",
                "--dirty",
                "--tags",
                "--long",
                "--abbrev=40",
                f"--match={tag_pattern}",
            ]
            result = subprocess.run(git_cmd, text=True, check=True, capture_output=True)  # noqa: S603
            describe_out = result.stdout.strip().split("-")

            git_cmd = ["git", "rev-parse", "--abbrev-ref", "HEAD"]
            result = subprocess.run(git_cmd, text=True, check=True, capture_output=True)  # noqa: S603
            branch_name = result.stdout.strip()
            short_branch_name = re.sub(r"([^a-zA-Z0-9]*)", "", branch_name).lower()[:20]
        except subprocess.CalledProcessError as e:
            logger.debug("Error when running git describe: %s", e.stderr)
            return SCMInfo(tool=cls)

        info = SCMInfo(tool=cls, branch_name=branch_name, short_branch_name=short_branch_name)

        if describe_out[-1].strip() == "dirty":
            info.dirty = True
            describe_out.pop()
        else:
            info.dirty = False

        info.commit_sha = describe_out.pop().lstrip("g")
        info.distance_to_latest_tag = int(describe_out.pop())
        version = cls.get_version_from_tag(describe_out[-1], tag_name, parse_pattern)
        info.current_version = version or "-".join(describe_out).lstrip("v")

        return info

    @classmethod
    def add_path(cls, path: Union[str, Path]) -> None:
        """Add a path to the VCS."""
        subprocess.check_output(["git", "add", "--update", str(path)])  # noqa: S603, S607

    @classmethod
    def tag(cls, name: str, sign: bool = False, message: Optional[str] = None) -> None:
        """
        Create a tag of the new_version in VCS.

        If only name is given, bumpversion uses a lightweight tag.
        Otherwise, it utilizes an annotated tag.

        Args:
            name: The name of the tag
            sign: True to sign the tag
            message: An optional message to annotate the tag.
        """
        command = ["git", "tag", name]
        if sign:
            command += ["--sign"]
        if message:
            command += ["--message", message]
        subprocess.check_output(command)  # noqa: S603


class Mercurial(SourceCodeManager):
    """Mercurial implementation."""

    _TEST_USABLE_COMMAND: ClassVar[List[str]] = ["hg", "root"]
    _COMMIT_COMMAND: ClassVar[List[str]] = ["hg", "commit", "--logfile"]
    _ALL_TAGS_COMMAND: ClassVar[List[str]] = ["hg", "log", '--rev="tag()"', '--template="{tags}\n"']

    @classmethod
    def latest_tag_info(cls, tag_name: str, parse_pattern: str) -> SCMInfo:
        """Return information about the latest tag."""
        current_version = None
        re_pattern = tag_name.replace("{new_version}", ".*")
        result = subprocess.run(
            ["hg", "log", "-r", f"tag('re:{re_pattern}')", "--template", "{latesttag}\n"],  # noqa: S603, S607
            text=True,
            check=True,
            capture_output=True,
        )
        result.check_returncode()
        if result.stdout:
            tag_string = result.stdout.splitlines(keepends=False)[-1]
            current_version = cls.get_version_from_tag(tag_string, tag_name, parse_pattern)
        else:
            logger.debug("No tags found")
        is_dirty = len(subprocess.check_output(["hg", "status", "-mard"])) != 0  # noqa: S603, S607
        return SCMInfo(tool=cls, current_version=current_version, dirty=is_dirty)

    @classmethod
    def assert_nondirty(cls) -> None:
        """Assert that the working directory is clean."""
        lines = [
            line.strip()
            for line in subprocess.check_output(["hg", "status", "-mard"]).splitlines()  # noqa: S603, S607
            if not line.strip().startswith(b"??")
        ]

        if lines:
            joined_lines = b"\n".join(lines).decode()
            raise DirtyWorkingDirectoryError(f"Mercurial working directory is not clean:\n{joined_lines}")

    @classmethod
    def add_path(cls, path: Union[str, Path]) -> None:
        """Add a path to the VCS."""
        pass

    @classmethod
    def tag(cls, name: str, sign: bool = False, message: Optional[str] = None) -> None:
        """
        Create a tag of the new_version in VCS.

        If only name is given, bumpversion uses a lightweight tag.
        Otherwise, it utilizes an annotated tag.

        Args:
            name: The name of the tag
            sign: True to sign the tag
            message: A optional message to annotate the tag.

        Raises:
            SignedTagsError: If ``sign`` is ``True``
        """
        command = ["hg", "tag", name]
        if sign:
            raise SignedTagsError("Mercurial does not support signed tags.")
        if message:
            command += ["--message", message]
        subprocess.check_output(command)  # noqa: S603


def get_scm_info(tag_name: str, parse_pattern: str) -> SCMInfo:
    """Return a dict with the latest source code management info."""
    if Git.is_usable():
        return Git.latest_tag_info(tag_name, parse_pattern)
    elif Mercurial.is_usable():
        return Mercurial.latest_tag_info(tag_name, parse_pattern)
    else:
        return SCMInfo()



"""This module handles the legacy config file format."""

from __future__ import annotations

import re
from difflib import context_diff
from pathlib import Path
from typing import Any, Dict, Union

from bumpversion.ui import get_indented_logger

logger = get_indented_logger(__name__)


def read_ini_file(file_path: Path) -> Dict[str, Any]:  # noqa: C901
    """
    Parse an INI file and return a dictionary of sections and their options.

    Args:
        file_path: The path to the INI file.

    Returns:
        dict: A dictionary of sections and their options.
    """
    import configparser

    from bumpversion import autocast

    # Create a ConfigParser object and read the INI file
    config_parser = configparser.RawConfigParser()
    if file_path.name == "setup.cfg":
        config_parser = configparser.ConfigParser()

    config_parser.read(file_path)

    # Create an empty dictionary to hold the parsed sections and options
    bumpversion_options: Dict[str, Any] = {"files": [], "parts": {}}

    # Loop through each section in the INI file
    for section_name in config_parser.sections():
        if not section_name.startswith("bumpversion"):
            continue

        section_parts = section_name.split(":")
        num_parts = len(section_parts)
        options = {key: autocast.autocast_value(val) for key, val in config_parser.items(section_name)}
        if "current_version" in options:
            options["current_version"] = str(options["current_version"])

        if num_parts == 1:  # bumpversion section
            bumpversion_options.update(options)
            serialize = bumpversion_options.get("serialize", [])
            if "message" in bumpversion_options and isinstance(bumpversion_options["message"], list):
                bumpversion_options["message"] = ",".join(bumpversion_options["message"])
            if not isinstance(serialize, list):
                bumpversion_options["serialize"] = [serialize]
        elif num_parts > 1 and section_parts[1].startswith("file"):
            file_options = {
                "filename": section_parts[2],
            }
            file_options.update(options)
            if "search" in file_options and isinstance(file_options["search"], list):
                file_options["search"] = "\n".join(file_options["search"])
            if "replace" in file_options and isinstance(file_options["replace"], list):
                file_options["replace"] = "\n".join(file_options["replace"])
            bumpversion_options["files"].append(file_options)
        elif num_parts > 1 and section_parts[1].startswith("glob"):
            file_options = {
                "glob": section_parts[2],
            }
            file_options.update(options)
            if "search" in file_options and isinstance(file_options["search"], list):
                file_options["search"] = "\n".join(file_options["search"])
            if "replace" in file_options and isinstance(file_options["replace"], list):
                file_options["replace"] = "\n".join(file_options["replace"])
            bumpversion_options["files"].append(file_options)
        elif num_parts > 1 and section_parts[1].startswith("part"):
            bumpversion_options["parts"][section_parts[2]] = options

    # Return the dictionary of sections and options
    return bumpversion_options


def update_ini_config_file(
    config_file: Union[str, Path], current_version: str, new_version: str, dry_run: bool = False
) -> None:
    """
    Update the current_version key in the configuration file.

    Instead of parsing and re-writing the config file with new information, it will use
    a regular expression to just replace the current_version value. The idea is it will
    avoid unintentional changes (like formatting) to the config file.

    Args:
        config_file: The configuration file to explicitly use.
        current_version: The serialized current version.
        new_version: The serialized new version.
        dry_run: True if the update should be a dry run.
    """
    cfg_current_version_regex = re.compile(
        f"(?P<section_prefix>\\[bumpversion]\n[^[]*current_version\\s*=\\s*)(?P<version>{current_version})",
        re.MULTILINE,
    )

    config_path = Path(config_file)
    existing_config = config_path.read_text(encoding="utf-8")
    if config_path.suffix == ".cfg" and cfg_current_version_regex.search(existing_config):
        sub_str = f"\\g<section_prefix>{new_version}"
        new_config = cfg_current_version_regex.sub(sub_str, existing_config)
    else:
        logger.info("Could not find the current version in the config file: %s.", config_path)
        return

    logger.info(
        "%s to config file %s:",
        "Would write" if dry_run else "Writing",
        config_path,
    )

    logger.info(
        "\n".join(
            list(
                context_diff(
                    existing_config.splitlines(),
                    new_config.splitlines(),
                    fromfile=f"before {config_path}",
                    tofile=f"after {config_path}",
                    lineterm="",
                )
            )
        )
    )

    if not dry_run:
        config_path.write_text(new_config, encoding="utf-8")



"""General utilities."""

import string
from typing import Any, List, Tuple


def extract_regex_flags(regex_pattern: str) -> Tuple[str, str]:
    """
    Extract the regex flags from the regex pattern.

    Args:
        regex_pattern: The pattern that might start with regex flags

    Returns:
        A tuple of the regex pattern without the flag string and regex flag string
    """
    import re

    flag_pattern = r"^(\(\?[aiLmsux]+\))"
    bits = re.split(flag_pattern, regex_pattern)
    return (regex_pattern, "") if len(bits) == 1 else (bits[2], bits[1])


def recursive_sort_dict(input_value: Any) -> Any:
    """Sort a dictionary recursively."""
    if not isinstance(input_value, dict):
        return input_value

    return {key: recursive_sort_dict(input_value[key]) for key in sorted(input_value.keys())}


def key_val_string(d: dict) -> str:
    """Render the dictionary as a comma-delimited key=value string."""
    return ", ".join(f"{k}={v}" for k, v in sorted(d.items()))


def labels_for_format(serialize_format: str) -> List[str]:
    """Return a list of labels for the given serialize_format."""
    return [item[1] for item in string.Formatter().parse(serialize_format) if item[1]]


def get_overrides(**kwargs) -> dict:
    """Return a dictionary containing only the overridden key-values."""
    return {key: val for key, val in kwargs.items() if val is not None}


def get_nested_value(d: dict, path: str) -> Any:
    """
    Retrieves the value of a nested key in a dictionary based on the given path.

    Args:
        d: The dictionary to search.
        path: A string representing the path to the nested key, separated by periods.

    Returns:
        The value of the nested key.

    Raises:
        KeyError: If a key in the path does not exist.
        ValueError: If an element in the path is not a dictionary.
    """
    keys = path.split(".")
    current_element = d

    for key in keys:
        if not isinstance(current_element, dict):
            raise ValueError(f"Element at '{'.'.join(keys[:keys.index(key)])}' is not a dictionary")

        if key not in current_element:
            raise KeyError(f"Key '{key}' not found at '{'.'.join(keys[:keys.index(key)])}'")

        current_element = current_element[key]

    return current_element


def set_nested_value(d: dict, value: Any, path: str) -> None:
    """
    Sets the value of a nested key in a dictionary based on the given path.

    Args:
        d: The dictionary to search.
        value: The value to set.
        path: A string representing the path to the nested key, separated by periods.

    Raises:
        ValueError: If an element in the path is not a dictionary.
    """
    keys = path.split(".")
    last_element = keys[-1]
    current_element = d

    for i, key in enumerate(keys):
        if key == last_element:
            current_element[key] = value
        elif key not in current_element:
            raise KeyError(f"Key '{key}' not found at '{'.'.join(keys[:keys.index(key)])}'")
        elif not isinstance(current_element[key], dict):
            raise ValueError(f"Path '{'.'.join(keys[:i+1])}' does not lead to a dictionary.")
        else:
            current_element = current_element[key]



import shutil
from dataclasses import dataclass
from pathlib import Path

from tests.conftest import get_config_data, inside_dir

import pytest
from pytest import param

from bumpversion import show
from bumpversion import config

mapping = {"key1": "value1", "dict-key": {"dict-key-key1": "value2-1"}}

FIXTURES_PATH = Path(__file__).parent.joinpath("fixtures")


@dataclass
class SimpleObj:
    """An object for testing."""

    dict_attr: dict
    list_attr: list

    @property
    def exc(self):
        return self._notavailable


test_obj = SimpleObj(dict_attr=mapping, list_attr=["a", "b", "c", "d"])


@pytest.mark.parametrize(
    ["name", "data", "expected"],
    [
        param("key1", mapping, "value1", id="simple mapping"),
        param("dict-key.dict-key-key1", mapping, "value2-1", id="nested mapping"),
        param("dict_attr", test_obj, mapping, id="attribute lookup"),
        param("list_attr.2", test_obj, "c", id="list lookup"),
    ],
)
def test_resolve_name(name: str, data, expected):
    """Test the resolve_name method gets the correct values."""
    assert show.resolve_name(data, name) == expected


def test_resolve_name_default():
    """Test a default value."""
    assert show.resolve_name(mapping, "key3", default="default") == "default"


def test_resolve_name_property_error():
    """An error in a property returns default."""
    assert show.resolve_name(test_obj, "exc", default="default") == "default"


DEFAULT_OUTPUT = FIXTURES_PATH.joinpath("basic_cfg_expected.txt").read_text().strip()
YAML_OUTPUT = FIXTURES_PATH.joinpath("basic_cfg_expected.yaml").read_text().strip()
JSON_OUTPUT = FIXTURES_PATH.joinpath("basic_cfg_expected_full.json").read_text().strip()

FOUR_ARGS_DEFAULT_OUTPUT = "{'allow_dirty': False,\n 'current_version': '1.0.0',\n 'sign_tags': False,\n 'tag': True}"
FOUR_ARGS_YAML_OUTPUT = 'allow_dirty: false\ncurrent_version: "1.0.0"\nsign_tags: false\ntag: true'
FOUR_ARGS_JSON_OUTPUT = '{"allow_dirty": false, "current_version": "1.0.0", "sign_tags": false, "tag": true}'

DOTTED_DEFAULT_OUTPUT = (
    "{'current_version': '1.0.0',\n 'files.1.filename': 'bumpversion/__init__.py',\n"
    " 'parts.release.values': ['dev', 'gamma']}"
)
DOTTED_YAML_OUTPUT = (
    'current_version: "1.0.0"\n'
    'files.1.filename: "bumpversion/__init__.py"\n'
    "parts.release.values:\n"
    '  - "dev"\n'
    '  - "gamma"'
)
DOTTED_JSON_OUTPUT = (
    '{\n  "current_version": "1.0.0",'
    '\n  "files.1.filename": "bumpversion/__init__.py",\n  "parts.release.values": [\n    "dev",\n    "gamma"\n  ]\n}'
)


@pytest.mark.parametrize(
    ["req_args", "format_", "expected"],
    (
        param(
            [],
            None,
            DEFAULT_OUTPUT,
            id="no-arguments-default",
        ),
        param(
            [],
            "yaml",
            YAML_OUTPUT,
            id="no-arguments-yaml",
        ),
        param(
            [],
            "json",
            JSON_OUTPUT,
            id="no-arguments-json",
        ),
        param(
            ["all"],
            None,
            DEFAULT_OUTPUT,
            id="all-argument-default",
        ),
        param(
            ["all", "current_version"],
            None,
            DEFAULT_OUTPUT,
            id="all-argument-in-args-default",
        ),
        param(
            ["current_version"],
            "default",
            "1.0.0",
            id="current_version",
        ),
        param(
            ["current_version", "allow_dirty", "sign_tags", "tag"],
            "default",
            FOUR_ARGS_DEFAULT_OUTPUT,
            id="four_args_default",
        ),
        param(
            ["files.1.filename", "current_version", "parts.release.values"],
            "default",
            DOTTED_DEFAULT_OUTPUT,
            id="dotted-default",
        ),
        param(
            ["files.1.filename", "current_version", "parts.release.values"],
            "yaml",
            DOTTED_YAML_OUTPUT,
            id="dotted-yaml",
        ),
        param(
            ["files.1.filename", "current_version", "parts.release.values"],
            "json",
            DOTTED_JSON_OUTPUT,
            id="dotted-json",
        ),
    ),
)
def test_do_show(
    req_args: list, format_: str, expected: str, tmp_path: Path, fixtures_path: Path, capsys: pytest.CaptureFixture
) -> None:
    """Test the show method."""
    config_path = tmp_path / "pyproject.toml"
    toml_path = fixtures_path / "basic_cfg.toml"
    shutil.copy(toml_path, config_path)
    with inside_dir(tmp_path):
        conf = config.get_configuration(config_file=fixtures_path.joinpath(config_path))
        show.do_show(config=conf, format_=format_, *req_args)
    captured = capsys.readouterr()
    assert captured.out.strip() == expected


def test_do_show_increment(tmp_path: Path, fixtures_path: Path, capsys: pytest.CaptureFixture) -> None:
    """Test the show method with increment."""
    config_path = tmp_path / "pyproject.toml"
    toml_path = fixtures_path / "basic_cfg.toml"
    shutil.copy(toml_path, config_path)
    with inside_dir(tmp_path):
        conf = config.get_configuration(config_file=fixtures_path.joinpath(config_path))
        show.do_show("current_version", "new_version", config=conf, format_="default", increment="minor")
    captured = capsys.readouterr()
    assert captured.out.strip() == "{'current_version': '1.0.0', 'new_version': '1.1.0-dev'}"



"""Generators for version parts."""

import datetime
import re
from typing import List, Optional, Union

from bumpversion.context import get_datetime_info


class PartFunction:
    """Base class for a version part function."""

    first_value: str
    optional_value: str
    independent: bool
    always_increment: bool

    def bump(self, value: str) -> str:
        """Increase the value."""
        raise NotImplementedError


class IndependentFunction(PartFunction):
    """
    This is a class that provides an independent function for version parts.

    It simply returns the optional value, which is equal to the first value.
    """

    def __init__(self, value: Union[str, int, None] = None):
        if value is None:
            value = ""
        self.first_value = str(value)
        self.optional_value = str(value)
        self.independent = True
        self.always_increment = False

    def bump(self, value: Optional[str] = None) -> str:
        """Return the optional value."""
        return value or self.optional_value


class CalVerFunction(PartFunction):
    """This is a class that provides a CalVer function for version parts."""

    def __init__(self, calver_format: str):
        self.independent = False
        self.calver_format = calver_format
        self.first_value = self.bump()
        self.optional_value = "There isn't an optional value for CalVer."
        self.independent = False
        self.always_increment = True

    def bump(self, value: Optional[str] = None) -> str:
        """Return the optional value."""
        return self.calver_format.format(**get_datetime_info(datetime.datetime.now()))


class NumericFunction(PartFunction):
    """
    This is a class that provides a numeric function for version parts.

    It simply starts with the provided first_value (0 by default) and
    increases it following the sequence of integer numbers.

    The optional value of this function is equal to the first value.

    This function also supports alphanumeric parts, altering just the numeric
    part (e.g. 'r3' --> 'r4'). Only the first numeric group found in the part is
    considered (e.g. 'r3-001' --> 'r4-001').
    """

    FIRST_NUMERIC = re.compile(r"(?P<prefix>[^-0-9]*)(?P<number>-?\d+)(?P<suffix>.*)")

    def __init__(self, optional_value: Union[str, int, None] = None, first_value: Union[str, int, None] = None):
        if first_value is not None and not self.FIRST_NUMERIC.search(str(first_value)):
            raise ValueError(f"The given first value {first_value} does not contain any digit")

        self.first_value = str(first_value or 0)
        self.optional_value = str(optional_value or self.first_value)
        self.independent = False
        self.always_increment = False

    def bump(self, value: Union[str, int]) -> str:
        """Increase the first numerical value by one."""
        match = self.FIRST_NUMERIC.search(str(value))
        if not match:
            raise ValueError(f"The given value {value} does not contain any digit")

        part_prefix, part_numeric, part_suffix = match.groups()

        if int(part_numeric) < int(self.first_value):
            raise ValueError(
                f"The given value {value} is lower than the first value {self.first_value} and cannot be bumped."
            )

        bumped_numeric = int(part_numeric) + 1

        return "".join([part_prefix, str(bumped_numeric), part_suffix])


class ValuesFunction(PartFunction):
    """
    This is a class that provides a values list based function for version parts.

    It is initialized with a list of values and iterates through them when
    bumping the part.

    The default optional value of this function is equal to the first value,
    but may be otherwise specified.

    When trying to bump a part which has already the maximum value in the list
    you get a ValueError exception.
    """

    def __init__(
        self,
        values: List[str],
        optional_value: Optional[str] = None,
        first_value: Optional[str] = None,
    ):
        if not values:
            raise ValueError("Version part values cannot be empty")

        self._values = values
        self.independent = False

        if optional_value is None:
            optional_value = values[0]

        if optional_value not in values:
            raise ValueError(f"Optional value {optional_value} must be included in values {values}")

        self.optional_value = optional_value

        if not first_value:
            first_value = values[0]

        if first_value not in values:
            raise ValueError(f"First value {first_value} must be included in values {values}")

        self.first_value = first_value

    def bump(self, value: str) -> str:
        """Return the item after ``value`` in the list."""
        try:
            return self._values[self._values.index(value) + 1]
        except IndexError as e:
            raise ValueError(
                f"The part has already the maximum value among {self._values} and cannot be bumped."
            ) from e



"""Tests the bump CLI subcommand."""

import shutil
import subprocess
import traceback
from datetime import datetime
from pathlib import Path

import pytest
from pytest import param

from click.testing import Result

from bumpversion import cli
from tests.conftest import inside_dir


class TestNoConfiguredFilesOption:
    """Tests around the behavior of the --no-configured-files option."""

    def test_no_files_marked_to_modify(self, mocker, tmp_path: Path, runner):
        """
        No files are sent to the `do_bump` function if the --no-configured-files option is used.
        """
        mocked_do_bump = mocker.patch("bumpversion.cli.do_bump")
        with inside_dir(tmp_path):
            result: Result = runner.invoke(
                cli.cli, ["bump", "--current-version", "1.0.0", "--no-configured-files", "patch"]
            )

        if result.exit_code != 0:
            print(result.output)

        assert result.exit_code == 0

        call_args = mocked_do_bump.call_args[0]
        passed_config = call_args[2]
        assert len(passed_config.files) == 0

    def test_file_args_marked_to_modify(self, mocker, tmp_path: Path, runner):
        """File paths marked in the command-line arguments are sent to the `do_bump` function."""
        mocked_do_bump = mocker.patch("bumpversion.cli.do_bump")
        with inside_dir(tmp_path):
            result: Result = runner.invoke(
                cli.cli,
                ["bump", "--current-version", "1.0.0", "--no-configured-files", "patch", "do-this-file.txt"],
            )

        if result.exit_code != 0:
            print(result.output)

        assert result.exit_code == 0

        call_args = mocked_do_bump.call_args[0]
        passed_config = call_args[2]
        assert len(passed_config.files) == 1
        assert passed_config.files[0].filename == "do-this-file.txt"


def test_bump_nested_regex(tmp_path: Path, fixtures_path: Path, caplog, runner):
    """
    Arrange/Act: Run the `bump` subcommand with --no-configured-files.

    Assert: There is no configured files specified to modify
    """
    cff_path = tmp_path / "citation.cff"
    cff_path.write_text("cff-version: 1.2.0\ndate-released: 2023-09-19\n", encoding="utf-8")
    content = fixtures_path.joinpath("regex_test_config.toml").read_text()
    config_path = tmp_path / ".bumpversion.toml"
    config_path.write_text(content, encoding="utf-8")

    with inside_dir(tmp_path):
        result: Result = runner.invoke(cli.cli, ["bump", "-vv", "patch"])

    if result.exit_code != 0:
        print(result.output)
        print(caplog.text)

    assert result.exit_code == 0

    now = datetime.now().isoformat()[:10]
    assert cff_path.read_text() == f"cff-version: 1.2.0\ndate-released: {now}\n"


def test_missing_explicit_config_file(tmp_path: Path, runner):
    """The command-line processor should raise an exception if the config file is missing."""
    with inside_dir(tmp_path):
        with inside_dir(tmp_path):
            result: Result = runner.invoke(cli.cli, ["bump", "--config-file", "missing-file.cfg"])
        assert result.exit_code != 0
        assert "'missing-file.cfg' does not exist." in result.output


def test_cli_options_override_config(tmp_path: Path, fixtures_path: Path, mocker, runner):
    """The command-line processor should override the config file."""
    # Arrange
    config_path = tmp_path / "this_config.toml"
    fixture_toml = fixtures_path / "basic_cfg.toml"
    shutil.copy(fixture_toml, config_path)
    mocked_do_bump = mocker.patch("bumpversion.cli.do_bump")

    # Act
    with inside_dir(tmp_path):
        result: Result = runner.invoke(
            cli.cli,
            [
                "bump",
                "--config-file",
                str(config_path),
                "--current-version",
                "1.1.0",
                "--new-version",
                "1.2.0",
                "--allow-dirty",
                "--parse",
                r"XXX(?P<spam>\d+);(?P<blob>\d+);(?P<slurp>\d+)",
                "--serialize",
                "XXX{spam};{blob};{slurp}",
                "--search",
                "my-search",
                "--replace",
                "my-replace",
                "--no-commit",
                "--no-tag",
                "slurp",
                "do-this-file.txt",
            ],
        )

    # Assert
    assert result.exit_code == 0
    assert mocked_do_bump.call_count == 1
    assert mocked_do_bump.call_args[0][0] == "slurp"
    assert mocked_do_bump.call_args[0][1] == "1.2.0"
    the_config = mocked_do_bump.call_args[0][2]
    assert the_config.current_version == "1.1.0"
    assert the_config.allow_dirty
    assert the_config.parse == r"XXX(?P<spam>\d+);(?P<blob>\d+);(?P<slurp>\d+)"
    assert the_config.serialize == ("XXX{spam};{blob};{slurp}",)
    assert the_config.search == "my-search"
    assert the_config.replace == "my-replace"
    assert the_config.commit is False
    assert the_config.tag is False
    assert len(the_config.files) == 4
    assert {f.filename for f in the_config.files} == {
        "setup.py",
        "bumpversion/__init__.py",
        "CHANGELOG.md",
        "do-this-file.txt",
    }
    assert mocked_do_bump.call_args[0][3] == config_path


@pytest.mark.parametrize(
    ["repo", "scm_command"],
    [
        param("git_repo", "git", id="git"),
        param("hg_repo", "hg", id="hg"),
    ],
)
def test_dirty_work_dir_raises_error(repo: str, scm_command: str, request, runner):
    repo_path: Path = request.getfixturevalue(repo)
    with inside_dir(repo_path):
        # Arrange
        repo_path.joinpath("dirty2").write_text("i'm dirty! 1.1.1", encoding="utf-8")
        subprocess.run([scm_command, "add", "dirty2"], check=True)

        # Act
        result: Result = runner.invoke(
            cli.cli, ["bump", "patch", "--current-version", "1.1.1", "--no-allow-dirty", "dirty2"]
        )

    # Assert
    assert result.exit_code != 0
    assert "working directory is not clean" in result.output


def test_non_scm_operations_if_scm_not_installed(tmp_path: Path, monkeypatch, runner):
    """Everything works except SCM commands if the SCM is unusable."""
    # Arrange
    monkeypatch.setenv("PATH", "")

    with inside_dir(tmp_path):
        version_path = tmp_path / "VERSION"
        version_path.write_text("31.0.3", encoding="utf-8")

        # Act
        runner.invoke(cli.cli, ["bump", "major", "--current-version", "31.0.3", "VERSION"])

    # Assert
    assert version_path.read_text() == "32.0.0"


@pytest.mark.parametrize(
    ["version_part"],
    [
        param("charlie", id="bad_version_part"),
        param("", id="missing_version_part"),
    ],
)
def test_detects_bad_or_missing_version_part(version_part: str, tmp_path: Path, monkeypatch, runner):
    """It properly detects bad or missing version part."""
    # Arrange
    monkeypatch.setenv("PATH", "")

    with inside_dir(tmp_path):
        version_path = tmp_path / "VERSION"
        version_path.write_text("31.0.3", encoding="utf-8")

        args = ["bump", "--current-version", "31.0.3"]
        if version_part:
            args.append(version_part)
        args.append("VERSION")
        # Act
        result = runner.invoke(cli.cli, args)

    # Assert
    assert result.exception is not None
    assert "Unknown version part:" in result.stdout


def test_ignores_missing_files_with_option(tmp_path, fixtures_path, runner):
    """The replace subcommand should ignore missing."""

    config_file = tmp_path / ".bumpversion.toml"
    config_file.write_text(
        "[tool.bumpversion]\n"
        'current_version = "0.0.1"\n'
        "allow_dirty = true\n\n"
        "[[tool.bumpversion.files]]\n"
        'filename = "VERSION"\n'
        "regex = false\n",
        encoding="utf-8",
    )

    # Act
    with inside_dir(tmp_path):
        result: Result = runner.invoke(
            cli.cli,
            [
                "bump",
                "--verbose",
                "--no-regex",
                "--no-configured-files",
                "--ignore-missing-files",
                "minor",
                "VERSION",
            ],
        )

    # Assert
    if result.exit_code != 0:
        print("Here is the output:")
        print(result.output)
        print(traceback.print_exception(result.exc_info[1]))

    assert result.exit_code == 0



import pytest

from bumpversion.indented_logger import IndentedLoggerAdapter
import logging


def test_does_not_indent_without_intent(caplog: pytest.LogCaptureFixture):
    caplog.set_level(logging.DEBUG)
    logger = IndentedLoggerAdapter(logging.getLogger(), reset=True)
    logger.debug("test debug")
    logger.info("test info")
    logger.warning("test warning")
    logger.error("test error")
    logger.critical("test critical")

    assert caplog.record_tuples == [
        ("root", 10, "test debug"),
        ("root", 20, "test info"),
        ("root", 30, "test warning"),
        ("root", 40, "test error"),
        ("root", 50, "test critical"),
    ]


def test_indents(caplog: pytest.LogCaptureFixture):
    caplog.set_level(logging.DEBUG)
    logger = IndentedLoggerAdapter(logging.getLogger(), reset=True)
    logger.info("test 1")
    logger.indent(2)
    logger.error("test %d", 2)
    logger.indent()
    logger.debug("test 3")
    logger.warning("test 4")
    logger.indent()
    logger.critical("test 5")
    logger.critical("test 6")

    assert caplog.record_tuples == [
        ("root", 20, "test 1"),
        ("root", 40, "    test 2"),
        ("root", 10, "      test 3"),
        ("root", 30, "      test 4"),
        ("root", 50, "        test 5"),
        ("root", 50, "        test 6"),
    ]


def test_dedents(caplog: pytest.LogCaptureFixture):
    caplog.set_level(logging.DEBUG)
    logger = IndentedLoggerAdapter(logging.getLogger(), reset=True)
    logger.indent(3)
    logger.info("test 1")
    logger.dedent(2)
    logger.error("test %d", 2)
    logger.dedent()
    logger.debug("test 3")
    logger.warning("test 4")

    assert caplog.record_tuples == [
        ("root", 20, "      test 1"),
        ("root", 40, "  test 2"),
        ("root", 10, "test 3"),
        ("root", 30, "test 4"),
    ]


def test_cant_dedent_below_zero(caplog: pytest.LogCaptureFixture):
    caplog.set_level(logging.DEBUG)
    logger = IndentedLoggerAdapter(logging.getLogger())
    logger.dedent(4)
    logger.info("test 1")

    assert caplog.record_tuples == [("root", 20, "test 1")]


def test_current_indent_shared_by_multiple_loggers(caplog: pytest.LogCaptureFixture):
    """Indenting one logger indents all loggers."""
    caplog.set_level(logging.DEBUG)
    logger1 = IndentedLoggerAdapter(logging.getLogger("logger1"), reset=True)
    logger1.info("test 1")
    logger1.indent()
    logger2 = IndentedLoggerAdapter(logging.getLogger("logger2"))
    logger2.info("test 2")
    logger3 = IndentedLoggerAdapter(logging.getLogger("logger3"))
    logger3.indent()
    logger3.info("test 3")
    logger3.info("test 4")
    logger2.dedent()
    logger2.info("test 5")

    assert caplog.record_tuples == [
        ("logger1", 20, "test 1"),
        ("logger2", 20, "  test 2"),
        ("logger3", 20, "    test 3"),
        ("logger3", 20, "    test 4"),
        ("logger2", 20, "  test 5"),
    ]



"""A simple YAML dumper to avoid extra dependencies."""

import datetime
from collections import UserDict
from io import StringIO
from textwrap import indent
from typing import Any, Callable, Union

DumperFunc = Callable[[Any], str]


class YAMLDumpers(UserDict):
    """Registry of YAML dumpers."""

    def add_dumper(self, data_type: type, dumper: DumperFunc) -> None:
        """Add a YAML dumper."""
        self.data[data_type] = dumper


YAML_DUMPERS = YAMLDumpers()

INDENT = "  "


def dump(data: Any) -> str:
    """Dump a value to a string buffer."""
    data_types = type(data).__mro__
    for data_type in data_types:
        if data_type in YAML_DUMPERS:
            return YAML_DUMPERS[data_types[0]](data)

    return YAML_DUMPERS[str](str(data))


def format_str(val: str) -> str:
    """Return a YAML representation of a string."""
    val = val.replace("\\", "\\\\").replace("\n", "\\n")
    return f'"{val}"'


YAML_DUMPERS.add_dumper(str, format_str)


def format_int(val: int) -> str:
    """Return a YAML representation of an int."""
    return str(val)


YAML_DUMPERS.add_dumper(int, format_int)


def format_float(data: float) -> str:
    """Return a YAML representation of a float."""
    inf_value = 1e300

    if data != data:
        value = ".nan"
    elif data == inf_value:
        value = ".inf"
    elif data == -inf_value:
        value = "-.inf"
    else:
        value = repr(data).lower()
        # Note that in some cases `repr(data)` represents a float number
        # without the decimal parts.  For instance:
        #   >>> repr(1e17)
        #   '1e17'
        # Unfortunately, this is not a valid float representation according
        # to the definition of the `!!float` tag.  We fix this by adding
        # '.0' before the 'e' symbol.
        if "." not in value and "e" in value:
            value = value.replace("e", ".0e", 1)
    return str(value)


YAML_DUMPERS.add_dumper(float, format_float)


def format_bool(val: bool) -> str:
    """Return a YAML representation of a bool."""
    return "true" if val else "false"


YAML_DUMPERS.add_dumper(bool, format_bool)


def format_dict(val: dict) -> str:
    """Return a YAML representation of a dict."""
    buffer = StringIO()

    for key, value in sorted(val.items()):
        rendered_value = dump(value).strip()
        if isinstance(value, (dict, list, tuple)):
            rendered_value = f"\n{indent(rendered_value, INDENT)}"
        else:
            rendered_value = f" {rendered_value}"
        buffer.write(f"{key}:{rendered_value}\n")

    return buffer.getvalue()


YAML_DUMPERS.add_dumper(dict, format_dict)


def format_sequence(val: Union[list, tuple]) -> str:
    """Return a string representation of a value."""
    buffer = StringIO()

    for item in val:
        rendered_value = dump(item).strip()
        if isinstance(item, dict):
            rendered_value = indent(rendered_value, INDENT).strip()

        if isinstance(item, (list, tuple)):
            rendered_value = f"\n{indent(rendered_value, INDENT)}"
        else:
            rendered_value = f" {rendered_value}"
        buffer.write(f"-{rendered_value}\n")

    return buffer.getvalue()


YAML_DUMPERS.add_dumper(list, format_sequence)
YAML_DUMPERS.add_dumper(tuple, format_sequence)


def format_none(_: None) -> str:
    """Return a YAML representation of None."""
    return "null"


YAML_DUMPERS.add_dumper(type(None), format_none)


def format_date(val: datetime.date) -> str:
    """Return a YAML representation of a date."""
    return val.isoformat()


YAML_DUMPERS.add_dumper(datetime.date, format_date)


def format_datetime(val: datetime.datetime) -> str:
    """Return a string representation of a value."""
    return val.isoformat(" ")


YAML_DUMPERS.add_dumper(datetime.datetime, format_datetime)



"""Tests for the yaml serialization module."""

from pathlib import Path

from bumpversion import yaml_dump
from datetime import datetime, date


def test_dump_unknown():
    assert yaml_dump.dump({1, 2}) == '"{1, 2}"'


def test_format_str():
    assert yaml_dump.format_str("value") == '"value"'


def test_format_int():
    assert yaml_dump.format_int(10) == "10"


def test_format_float():
    assert yaml_dump.format_float(1.0) == "1.0"
    assert yaml_dump.format_float(1e300) == ".inf"
    assert yaml_dump.format_float(-1e300) == "-.inf"
    assert yaml_dump.format_float(1e17) == "1.0e+17"
    assert yaml_dump.format_float(float("nan")) == ".nan"


def test_format_bool():
    assert yaml_dump.format_bool(True) == "true"
    assert yaml_dump.format_bool(False) == "false"


def test_format_dict():
    test_dict = {
        "key": "strval",
        "key2": 30,
        "key3": datetime(2023, 6, 19, 13, 45, 30),
        "key4": date(2023, 6, 19),
        "key5": {"subkey": "subval"},
        "key6": [1, 2, 3],
        "key7": None,
        "key8": True,
        "key9": False,
        "key10": 1.43,
        "key11": (1, 2, 3),
    }
    expected = (
        'key: "strval"\n'
        "key10: 1.43\n"
        "key11:\n"
        "  - 1\n"
        "  - 2\n"
        "  - 3\n"
        "key2: 30\n"
        "key3: 2023-06-19 13:45:30\n"
        "key4: 2023-06-19\n"
        "key5:\n"
        '  subkey: "subval"\n'
        "key6:\n"
        "  - 1\n"
        "  - 2\n"
        "  - 3\n"
        "key7: null\n"
        "key8: true\n"
        "key9: false\n"
    )
    assert yaml_dump.format_dict(test_dict) == expected


def test_format_list():
    assert yaml_dump.format_sequence(["item"]) == '- "item"\n'
    assert yaml_dump.format_sequence(["item", ["item2"]]) == '- "item"\n-\n  - "item2"\n'
    assert yaml_dump.format_sequence(("item",)) == '- "item"\n'
    assert yaml_dump.format_sequence(("item", ("item2",))) == '- "item"\n-\n  - "item2"\n'


def test_dump_none_val():
    assert yaml_dump.format_none(None) == "null"


def test_dump_date_val():
    test_date = date(2023, 6, 19)
    assert yaml_dump.format_date(test_date) == "2023-06-19"


def test_dump_datetime_val():
    test_datetime = datetime(2023, 6, 19, 13, 45, 30)
    assert yaml_dump.format_datetime(test_datetime) == "2023-06-19 13:45:30"



"""File processing tests."""

import logging
import os
import shutil
from datetime import datetime, timezone
from difflib import context_diff
from pathlib import Path
from textwrap import dedent
import tomlkit

import pytest
from pytest import param

from bumpversion import exceptions, files, config, bump
from bumpversion.config.models import FileChange
from bumpversion.context import get_context
from bumpversion.exceptions import VersionNotFoundError
from bumpversion.ui import setup_logging, get_indented_logger
from bumpversion.version_part import VersionConfig
from tests.conftest import get_config_data, inside_dir


class TestConfiguredFile:
    """Tests of the ConfiguredFile class."""

    class TestGetFileContents:
        """Tests of the get_file_content method."""

        def test_returns_contents_of_existing_file(self, tmp_path: Path) -> None:
            """The get_file_content method returns the contents of an existing file."""
            # Arrange
            filepath = tmp_path / "file.md"
            expected_data = "My current version is 1.2.3"
            filepath.write_text(expected_data, encoding="utf-8")
            overrides = {
                "current_version": "1.2.3",
                "files": [{"filename": str(filepath)}],
            }
            conf, version_config, current_version = get_config_data(overrides)
            assert len(conf.files_to_modify) == 1
            assert conf.files_to_modify[0].filename == str(filepath)

            # Act
            configured_file = files.ConfiguredFile(conf.files_to_modify[0], version_config)

            # Assert
            assert configured_file.get_file_contents() == expected_data

        def test_raises_exception_if_file_not_found(self, tmp_path: Path) -> None:
            """The get_file_content method raises an exception if the file does not exist."""
            filepath = tmp_path / "file.md"
            overrides = {
                "current_version": "1.2.3",
                "files": [{"filename": str(filepath)}],
            }
            conf, version_config, current_version = get_config_data(overrides)
            assert len(conf.files_to_modify) == 1
            assert conf.files_to_modify[0].filename == str(filepath)
            configured_file = files.ConfiguredFile(conf.files_to_modify[0], version_config)

            # Act
            with pytest.raises(FileNotFoundError):
                configured_file.get_file_contents()

    class TestMakeFileChange:
        """Tests of the make_file_change function."""

        def test_raises_exception_if_file_not_found(self, tmp_path: Path) -> None:
            """The make_file_change method raises an exception if the file does not exist."""
            # Assemble
            filepath = tmp_path / "file.md"
            overrides = {
                "current_version": "1.2.3",
                "files": [{"filename": str(filepath)}],
            }
            conf, version_config, current_version = get_config_data(overrides)
            configured_file = files.ConfiguredFile(conf.files_to_modify[0], version_config)
            new_version = current_version.bump("patch")
            ctx = get_context(conf)

            # Act
            with pytest.raises(FileNotFoundError):
                configured_file.make_file_change(current_version, new_version, ctx, dry_run=True)

        def test_logs_missing_file_when_ignore_missing_file_set(self, tmp_path: Path, caplog) -> None:
            """The make_file_change method logs missing file when ignore_missing_file set to True."""
            # Assemble
            setup_logging(1)
            caplog.set_level(logging.INFO)
            logger = get_indented_logger(__name__)
            logger.reset()
            filepath = tmp_path / "file.md"
            overrides = {
                "current_version": "1.2.3",
                "files": [{"filename": str(filepath), "ignore_missing_file": True}],
            }
            conf, version_config, current_version = get_config_data(overrides)
            configured_file = files.ConfiguredFile(conf.files_to_modify[0], version_config)
            new_version = current_version.bump("patch")
            ctx = get_context(conf)

            # Act
            configured_file.make_file_change(current_version, new_version, ctx, dry_run=True)

            # Assert
            logs = caplog.messages
            assert logs[0] == "Reading configuration"
            assert logs[1] == "  Configuration file not found: missing."
            assert logs[2] == f"\nFile {filepath}: replace `{{current_version}}` with `{{new_version}}`"
            assert logs[3] == "  File not found, but ignoring"

        def test_dedents_properly_when_file_does_not_contain_pattern(self, fixtures_path: Path, caplog) -> None:
            """The make_file_change method dedents the logging context when it does not contain a pattern."""
            # Assemble
            setup_logging(1)
            caplog.set_level(logging.INFO)
            logger = get_indented_logger(__name__)
            logger.reset()
            globs = fixtures_path / "glob" / "**/*.txt"
            overrides = {
                "current_version": "1.2.3",
                "files": [{"glob": str(globs), "ignore_missing_file": True, "ignore_missing_version": True}],
            }
            conf, version_config, current_version = get_config_data(overrides)
            configured_file1 = files.ConfiguredFile(conf.files_to_modify[0], version_config)
            configured_file2 = files.ConfiguredFile(conf.files_to_modify[1], version_config)
            configured_file3 = files.ConfiguredFile(conf.files_to_modify[2], version_config)
            new_version = current_version.bump("patch")
            ctx = get_context(conf)

            # Act
            configured_file1.make_file_change(current_version, new_version, ctx, dry_run=True)
            configured_file2.make_file_change(current_version, new_version, ctx, dry_run=True)
            configured_file3.make_file_change(current_version, new_version, ctx, dry_run=True)

            # Assert
            logs = caplog.messages
            assert logs[0] == "Reading configuration"
            assert logs[1] == "  Configuration file not found: missing."
            assert logs[2] == (
                f"\nFile {configured_file1.file_change.filename}: replace `{{current_version}}` with `{{new_version}}`"
            )
            assert logs[3] == (
                f"\nFile {configured_file2.file_change.filename}: replace `{{current_version}}` with `{{new_version}}`"
            )
            assert logs[4] == (
                f"\nFile {configured_file3.file_change.filename}: replace `{{current_version}}` with `{{new_version}}`"
            )


def test_single_file_processed_twice(tmp_path: Path):
    """
    Verify that a single file "file2" can be processed twice.

    Use two file_ entries, both with a different suffix after
    the underscore.
    Employ different parse/serialize and search/replace configs
    to verify correct interpretation.
    """
    filepath = tmp_path.joinpath("file2")
    filepath.write_text("dots: 0.10.2\ndashes: 0-10-2", encoding="utf-8")
    overrides = {
        "current_version": "0.10.2",
        "files": [
            {
                "filename": str(filepath),
                "search": "dots: {current_version}",
                "replace": "dots: {new_version}",
            },
            {
                "filename": str(filepath),
                "search": "dashes: {current_version}",
                "replace": "dashes: {new_version}",
                "parse": r"(?P<major>\d+)-(?P<minor>\d+)-(?P<patch>\d+)",
                "serialize": ["{major}-{minor}-{patch}"],
            },
        ],
    }
    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("patch")
    ctx = get_context(conf)

    assert len(conf.files) == 2
    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.make_file_change(current_version, new_version, ctx)

    assert filepath.read_text() == "dots: 0.10.3\ndashes: 0-10-3"


def test_multi_file_configuration(tmp_path: Path):
    full_vers_path = tmp_path / "FULL_VERSION.txt"
    full_vers_path.write_text("1.0.3", encoding="utf-8")
    maj_vers_path = tmp_path / "MAJOR_VERSION.txt"
    maj_vers_path.write_text("1", encoding="utf-8")
    readme_path = tmp_path / "README.txt"
    readme_path.write_text("MyAwesomeSoftware(TM) v1.0", encoding="utf-8")
    build_num_path = tmp_path / "BUILD_NUMBER"
    build_num_path.write_text("1.0.3+joe+38943", encoding="utf-8")

    overrides = {
        "current_version": "1.0.3+joe+38943",
        "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(?:\+.*)?",
        "serialize": [
            "{major}.{minor}.{patch}+{$USER}+{$BUILD_NUMBER}",
            "{major}.{minor}.{patch}",
        ],
        "files": [
            {
                "filename": str(full_vers_path),
            },
            {
                "filename": str(maj_vers_path),
                "parse": r"\d+",
                "serialize": ["{major}"],
            },
            {
                "filename": str(readme_path),
                "parse": r"(?P<major>\d+)\.(?P<minor>\d+)",
                "serialize": ["{major}.{minor}"],
            },
            {
                "filename": str(build_num_path),
                "serialize": ["{major}.{minor}.{patch}+{$USER}+{$BUILD_NUMBER}"],
            },
        ],
    }
    conf, version_config, current_version = get_config_data(overrides)
    major_version = current_version.bump("major")

    os.environ["BUILD_NUMBER"] = "38944"
    os.environ["USER"] = "bob"
    ctx = get_context(conf)
    del os.environ["BUILD_NUMBER"]
    del os.environ["USER"]

    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.make_file_change(current_version, major_version, ctx)

    assert full_vers_path.read_text() == "2.0.0"
    assert maj_vers_path.read_text() == "2"
    assert readme_path.read_text() == "MyAwesomeSoftware(TM) v2.0"
    assert build_num_path.read_text() == "2.0.0+bob+38944"

    os.environ["BUILD_NUMBER"] = "38945"
    os.environ["USER"] = "jane"
    conf.current_version = "2.0.0+bob+38944"
    ctx = get_context(conf)
    del os.environ["BUILD_NUMBER"]
    del os.environ["USER"]
    major_version = version_config.parse(conf.current_version)
    major_patch_version = major_version.bump("patch")
    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.make_file_change(major_version, major_patch_version, ctx)

    assert full_vers_path.read_text() == "2.0.1"
    assert maj_vers_path.read_text() == "2"
    assert readme_path.read_text() == "MyAwesomeSoftware(TM) v2.0"
    assert build_num_path.read_text() == "2.0.1+jane+38945"


def test_raises_correct_missing_version_string(tmp_path: Path, fixtures_path: Path):
    """When a file is missing the version string, the error should indicate the correct serialization missing."""
    csharp_path = fixtures_path / "csharp"
    dst_path: Path = shutil.copytree(csharp_path, tmp_path / "csharp")

    with inside_dir(dst_path):
        conf = config.get_configuration(config_file=dst_path.joinpath(".bumpversion.toml"))
        version_config = VersionConfig(conf.parse, conf.serialize, conf.search, conf.replace, conf.parts)
        current_version = version_config.parse(conf.current_version)
        dst_path.joinpath("Version.csv").write_text("1;3-1;0;rc;build.1031", encoding="utf-8")

        major_version = current_version.bump("patch")
        ctx = get_context(conf)

        configured_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

        with pytest.raises(VersionNotFoundError) as e:
            files.modify_files(configured_files, current_version, major_version, ctx)
            assert e.message.startswith("Did not find '1;3;1;0;rc;build.1031'")


def test_uses_correct_serialization_for_tag_and_commit_messages(git_repo: Path, fixtures_path: Path, caplog):
    """The tag and commit messages should use the root configured serialization."""
    import subprocess
    import logging

    caplog.set_level(logging.INFO)

    csharp_path = fixtures_path / "csharp"
    dst_path: Path = shutil.copytree(csharp_path, git_repo / "csharp")
    subprocess.check_call(["git", "add", "."], cwd=git_repo)
    subprocess.check_call(["git", "commit", "-m", "Initial commit"], cwd=git_repo)
    subprocess.check_call(["git", "tag", "3.1.0-rc+build.1031"], cwd=git_repo)

    with inside_dir(dst_path):
        config_file = dst_path.joinpath(".bumpversion.toml")
        conf = config.get_configuration(config_file=config_file)
        bump.do_bump("patch", None, conf, config_file, dry_run=True)

    assert "Bump version: 3.1.0-rc+build.1031 -> 3.1.1" in caplog.text


def test_search_replace_to_avoid_updating_unconcerned_lines(tmp_path: Path, caplog):
    import logging

    caplog.set_level(logging.DEBUG)
    req_path = tmp_path / "requirements.txt"
    req_path.write_text("Django>=1.5.6,<1.6\nMyProject==1.5.6", encoding="utf-8")

    changelog_path = tmp_path / "CHANGELOG.md"
    changelog_path.write_text(
        dedent(
            """
    # https://keepachangelog.com/en/1.0.0/

    ## [Unreleased]
    ### Added
    - Foobar

    ## [0.0.1] - 2014-05-31
    ### Added
    - This CHANGELOG file to hopefully serve as an evolving example of a
      standardized open source project CHANGELOG.
    """
        ),
        encoding="utf-8",
    )

    overrides = {
        "current_version": "1.5.6",
        "files": [
            {
                "filename": str(req_path),
                "search": "MyProject=={current_version}",
                "replace": "MyProject=={new_version}",
            },
            {
                "filename": str(changelog_path),
                "search": "## [Unreleased]",
                "replace": "## [Unreleased]\n\n## [{new_version}] - {utcnow:%Y-%m-%d}",
                "no_regex": True,
            },
        ],
    }

    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("minor")

    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.make_file_change(current_version, new_version, get_context(conf))

    utc_today = datetime.now(timezone.utc).strftime("%Y-%m-%d")
    expected_chglog = dedent(
        f"""
    # https://keepachangelog.com/en/1.0.0/

    ## [Unreleased]

    ## [1.6.0] - {utc_today}
    ### Added
    - Foobar

    ## [0.0.1] - 2014-05-31
    ### Added
    - This CHANGELOG file to hopefully serve as an evolving example of a
      standardized open source project CHANGELOG.
    """
    )
    assert req_path.read_text() == "Django>=1.5.6,<1.6\nMyProject==1.6.0"
    assert changelog_path.read_text() == expected_chglog


def test_non_matching_search_does_not_modify_file(tmp_path: Path):
    changelog_path = tmp_path / "CHANGELOG.md"
    changelog_content = dedent(
        """
    # Unreleased

    * bullet point A

    # Release v'older' (2019-09-17)

    * bullet point B
    """
    )
    changelog_path.write_text(changelog_content, encoding="utf-8")

    overrides = {
        "current_version": "1.0.3",
        "files": [
            {
                "filename": str(changelog_path),
                "search": "Not-yet-released",
                "replace": "Release v{new_version} ({now:%Y-%m-%d})",
            }
        ],
    }
    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("patch")
    configured_files = files.resolve_file_config(conf.files, version_config)
    with pytest.raises(exceptions.VersionNotFoundError, match="Did not find 'Not-yet-released' in file:"):
        files.modify_files(configured_files, current_version, new_version, get_context(conf))

    assert changelog_path.read_text() == changelog_content


def test_simple_replacement_in_utf8_file(tmp_path: Path):
    """Changing a file in UTF-8 should not change the non-ASCII characters."""
    # Arrange
    version_path = tmp_path / "VERSION"
    version_path.write_bytes("Kröt1.3.0".encode("utf-8"))

    overrides = {"current_version": "1.3.0", "files": [{"filename": str(version_path)}]}
    with inside_dir(tmp_path):
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("patch")

    # Act
    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.make_file_change(current_version, new_version, get_context(conf))

    # Assert
    out = version_path.read_text()
    assert out == "Kröt1.3.1"


def test_multi_line_search_is_found(tmp_path: Path) -> None:
    """A multiline search string is found and replaced."""
    # Arrange
    alphabet_path = tmp_path / "the_alphabet.txt"
    alphabet_path.write_text("A\nB\nC\n", encoding="utf-8")

    overrides = {
        "current_version": "9.8.7",
        "search": "A\nB\nC",
        "replace": "A\nB\nC\n{new_version}",
        "files": [{"filename": str(alphabet_path)}],
    }
    with inside_dir(tmp_path):
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("major")

    # Act
    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.make_file_change(current_version, new_version, get_context(conf))

    # Assert
    assert alphabet_path.read_text() == "A\nB\nC\n10.0.0\n"


@pytest.mark.parametrize(
    ["global_value", "file_value", "should_raise"],
    [
        param(True, True, False, id="ignore global and file"),
        param(True, False, True, id="ignore global only"),
        param(False, True, False, id="ignore file only"),
        param(False, False, True, id="ignore none"),
    ],
)
def test_ignore_missing_version(global_value: bool, file_value: bool, should_raise: bool, tmp_path: Path) -> None:
    """If the version is not found in the file, do nothing."""
    # Arrange
    version_path = tmp_path / Path("VERSION")
    version_path.write_text("1.2.3", encoding="utf-8")

    overrides = {
        "current_version": "1.2.5",
        "ignore_missing_version": global_value,
        "files": [{"filename": str(version_path), "ignore_missing_version": file_value}],
    }
    with inside_dir(tmp_path):
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("patch")
        cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

        # Act
        if should_raise:
            with pytest.raises(VersionNotFoundError):
                files.modify_files(cfg_files, current_version, new_version, get_context(conf))
        else:
            files.modify_files(cfg_files, current_version, new_version, get_context(conf))

    # Assert
    assert version_path.read_text() == "1.2.3"


def test_regex_search(tmp_path: Path) -> None:
    """A regex search string is found and replaced."""
    # Arrange
    version_path = tmp_path / "VERSION"
    version_path.write_text("Release: 1234-56-78 '1.2.3'", encoding="utf-8")

    overrides = {
        "regex": True,
        "current_version": "1.2.3",
        "search": r"Release: \d{{4}}-\d{{2}}-\d{{2}} '{current_version}'",
        "replace": r"Release {now:%Y-%m-%d} '{new_version}'",
        "files": [{"filename": str(version_path)}],
    }
    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("patch")
    cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

    # Act
    files.modify_files(cfg_files, current_version, new_version, get_context(conf))

    # Assert
    now = datetime.now().isoformat()[:10]
    assert version_path.read_text() == f"Release {now} '1.2.4'"


def test_regex_search_with_escaped_chars(tmp_path: Path) -> None:
    """A search that uses special characters is not treated as a regex."""
    # Arrange
    version_path = tmp_path / "VERSION"
    version_path.write_text("## [Release] 1.2.3 1234-56-78", encoding="utf-8")

    overrides = {
        "regex": True,
        "current_version": "1.2.3",
        "search": r"## \[Release\] {current_version} \d{{4}}-\d{{2}}-\d{{2}}",
        "replace": r"## [Release] {new_version} {now:%Y-%m-%d}",
        "files": [{"filename": str(version_path)}],
    }
    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("patch")
    cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

    # Act
    files.modify_files(cfg_files, current_version, new_version, get_context(conf))

    # Assert
    now = datetime.now().isoformat()[:10]
    assert version_path.read_text() == f"## [Release] 1.2.4 {now}"


def test_regex_search_in_toml(tmp_path: Path, fixtures_path: Path) -> None:
    """Tests how-to doc 'update-a-date.md'."""
    # Arrange
    version_path = tmp_path / "VERSION"
    version_path.write_text("__date__ = '1234-56-78'\n__version__ = '1.2.3'", encoding="utf-8")
    config_path = tmp_path / ".bumpversion.toml"
    shutil.copyfile(fixtures_path / "replace-date-config.toml", config_path)
    with inside_dir(tmp_path):
        conf = config.get_configuration(config_file=config_path)
        version_config = VersionConfig(conf.parse, conf.serialize, conf.search, conf.replace, conf.parts)
        current_version = version_config.parse(conf.current_version)

        new_version = current_version.bump("patch")
        cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

        # Act
        files.modify_files(cfg_files, current_version, new_version, get_context(conf))

    # Assert
    now = datetime.now().isoformat()[:10]
    assert version_path.read_text() == f"__date__ = '{now}'\n__version__ = '1.2.4'"


def test_regex_search_with_caret(tmp_path: Path, fixtures_path: Path) -> None:
    """A search that uses a caret to indicate the beginning of the line works correctly."""
    # Arrange
    config_path = tmp_path / ".bumpversion.toml"
    thingy_path = tmp_path / "thingy.yaml"
    shutil.copyfile(fixtures_path / "regex_with_caret.yaml", thingy_path)
    shutil.copyfile(fixtures_path / "regex_with_caret_config.toml", config_path)

    conf = config.get_configuration(config_file=config_path)
    version_config = VersionConfig(conf.parse, conf.serialize, conf.search, conf.replace, conf.parts)
    current_version = version_config.parse(conf.current_version)
    new_version = current_version.bump("patch")

    with inside_dir(tmp_path):
        cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

        # Act
        files.modify_files(cfg_files, current_version, new_version, get_context(conf))

    # Assert
    assert (
        thingy_path.read_text() == "version: 1.0.1\ndependencies:\n- name: kube-prometheus-stack\n  version: 1.0.0\n"
    )


def test_bad_regex_search(tmp_path: Path, caplog) -> None:
    """A search string not meant to be a regex is still found and replaced."""
    # Arrange
    version_path = tmp_path / "VERSION"
    version_path.write_text("Score: A+ ( '1.2.3'", encoding="utf-8")

    overrides = {
        "regex": True,
        "current_version": "1.2.3",
        "search": r"Score: A+ ( '{current_version}'",
        "replace": r"Score: A+ ( '{new_version}'",
        "files": [{"filename": str(version_path)}],
    }
    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("patch")
    cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

    # Act
    files.modify_files(cfg_files, current_version, new_version, get_context(conf))

    # Assert
    assert version_path.read_text() == "Score: A+ ( '1.2.4'"
    assert "Invalid regex" in caplog.text


class TestDataFileUpdater:
    """Tests for the DataFileUpdater class."""

    def test_update_file_does_not_modify_non_toml_files(self, tmp_path: Path) -> None:
        """A non-TOML file is not modified."""
        # Arrange
        version_path = tmp_path / "VERSION"
        version_path.write_text("1.2.3", encoding="utf-8")

        overrides = {"current_version": "1.2.3", "files": [{"filename": str(version_path)}]}
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("patch")
        datafile_config = FileChange(
            filename=str(version_path),
            key_path="",
            search=conf.search,
            replace=conf.replace,
            regex=conf.regex,
            ignore_missing_version=conf.ignore_missing_version,
            ignore_missing_file=conf.ignore_missing_files,
            serialize=conf.serialize,
            parse=conf.parse,
        )

        # Act
        files.DataFileUpdater(datafile_config, version_config.part_configs).update_file(
            current_version, new_version, get_context(conf)
        )

        # Assert
        assert version_path.read_text() == "1.2.3"

    def test_update_replaces_key(self, tmp_path: Path, fixtures_path: Path) -> None:
        """A key specific key is replaced and nothing else is touched."""
        # Arrange
        config_path = tmp_path / "pyproject.toml"
        fixture_path = fixtures_path / "partial_version_strings.toml"
        shutil.copy(fixture_path, config_path)

        contents_before = config_path.read_text()
        conf = config.get_configuration(config_file=config_path, files=[{"filename": str(config_path)}])
        version_config = VersionConfig(conf.parse, conf.serialize, conf.search, conf.replace, conf.parts)
        current_version = version_config.parse(conf.current_version)
        new_version = current_version.bump("minor")
        datafile_config = FileChange(
            filename=str(config_path),
            key_path="tool.bumpversion.current_version",
            search=conf.search,
            replace=conf.replace,
            regex=conf.regex,
            ignore_missing_version=conf.ignore_missing_version,
            ignore_missing_file=conf.ignore_missing_files,
            serialize=conf.serialize,
            parse=conf.parse,
        )

        # Act
        files.DataFileUpdater(datafile_config, version_config.part_configs).update_file(
            current_version, new_version, get_context(conf)
        )

        # Assert
        contents_after = config_path.read_text()
        toml_data = tomlkit.parse(config_path.read_text()).unwrap()
        actual_difference = list(
            context_diff(
                contents_before.splitlines(),
                contents_after.splitlines(),
                fromfile="before",
                tofile="after",
                n=0,
                lineterm="",
            )
        )
        expected_difference = [
            "*** before",
            "--- after",
            "***************",
            "*** 28 ****",
            '! current_version = "0.0.2"',
            "--- 28 ----",
            '! current_version = "0.1.0"',
        ]
        assert actual_difference == expected_difference
        assert toml_data["tool"]["pdm"]["dev-dependencies"]["lint"] == ["ruff==0.0.292"]
        assert toml_data["tool"]["bumpversion"]["current_version"] == "0.1.0"



import difflib
import json
from pathlib import Path
from textwrap import dedent

import pytest
from _pytest.mark import param

from bumpversion.config.files_legacy import read_ini_file, update_ini_config_file
from bumpversion import config
from ..conftest import inside_dir


@pytest.fixture(
    params=[
        "file",
        "file(suffix)",
        "file (suffix with space)",
        "file (suffix lacking closing paren",
    ]
)
def cfg_file_keyword(request):
    """Return multiple possible styles for the bumpversion:file keyword."""
    return request.param


@pytest.fixture(params=[".bumpversion.cfg", "setup.cfg"])
def cfg_file(request) -> str:
    """Return both config-file styles ('.bumpversion.cfg', 'setup.cfg')."""
    return request.param


@pytest.mark.parametrize(
    ["conf_file", "expected_file"],
    [
        param("basic_cfg.cfg", "basic_cfg_expected.json", id="ini basic cfg"),
        param("legacy_multiline_search.cfg", "legacy_multiline_search_expected.json", id="multiline search cfg"),
        param(
            "legacy_multiline_search_comma.cfg",
            "legacy_multiline_search_comma_expected.json",
            id="multiline search comma cfg",
        ),
    ],
)
def test_read_ini_file(conf_file: str, expected_file: str, fixtures_path: Path) -> None:
    """Parsing the config file should match the expected results."""
    result = read_ini_file(fixtures_path.joinpath(conf_file))
    expected = json.loads(fixtures_path.joinpath(expected_file).read_text())
    assert result == expected


def test_independent_falsy_value_in_config_does_not_bump_independently(tmp_path: Path):
    # tmp_path.joinpath("VERSION").write_text("2.1.0-5123")
    config_file = tmp_path.joinpath(".bumpversion.cfg")
    config_file.write_text(
        dedent(
            r"""
        [bumpversion]
        current_version: 2.1.0-5123
        parse = (?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)\-(?P<build>\d+)
        serialize = {major}.{minor}.{patch}-{build}

        [bumpversion:file:VERSION]

        [bumpversion:part:build]
        independent = 0
        """
        ),
        encoding="utf-8",
    )

    conf = config.get_configuration(config_file)
    assert conf.parts["build"].independent is False


def test_correct_interpolation_for_setup_cfg_files(tmp_path: Path, fixtures_path: Path):
    """
    Reported here: https://github.com/c4urself/bump2version/issues/21.
    """
    test_fixtures_path = fixtures_path.joinpath("interpolation")
    setup_cfg = config.get_configuration(test_fixtures_path.joinpath("setup.cfg"))
    bumpversion_cfg = config.get_configuration(test_fixtures_path.joinpath(".bumpversion.cfg"))
    pyproject_toml = config.get_configuration(test_fixtures_path.joinpath("pyproject.toml"))

    assert setup_cfg.replace == "{now:%m-%d-%Y} v. {new_version}"
    assert bumpversion_cfg.replace == "{now:%m-%d-%Y} v. {new_version}"
    assert pyproject_toml.replace == "{now:%m-%d-%Y} v. {new_version}"


CFG_EXPECTED_DIFF = (
    "*** \n"
    "--- \n"
    "***************\n"
    "*** 11 ****\n"
    "! current_version = 1.0.0\n"
    "--- 11 ----\n"
    "! current_version = 1.0.1\n"
)


@pytest.mark.parametrize(
    [
        "cfg_file_name",
    ],
    [
        (".bumpversion.cfg",),
        ("setup.cfg",),
    ],
)
def test_update_ini_config_file(tmp_path: Path, cfg_file_name: str, fixtures_path: Path) -> None:
    """
    Make sure only the version string is updated in the config file.
    """
    expected_diff = CFG_EXPECTED_DIFF
    cfg_path = tmp_path / cfg_file_name
    orig_path = fixtures_path / f"basic_cfg{cfg_path.suffix}"
    cfg_path.write_text(orig_path.read_text(), encoding="utf-8")
    original_content = orig_path.read_text().splitlines(keepends=True)

    update_ini_config_file(cfg_path, "1.0.0", "1.0.1")
    new_content = cfg_path.read_text().splitlines(keepends=True)
    difference = difflib.context_diff(original_content, new_content, n=0)
    assert "".join(difference) == expected_diff


def test_file_keyword_with_suffix_is_accepted(tmp_path: Path, cfg_file: str, cfg_file_keyword: str):
    cfg_file_path = tmp_path / cfg_file
    cfg_file_path.write_text(
        "[bumpversion]\n"
        "current_version = 0.10.2\n"
        "new_version = 0.10.3\n"
        "[bumpversion:file (foobar):file2]\n"
        "search = version {current_version}\n"
        "replace = version {new_version}\n"
        f"[bumpversion:{cfg_file_keyword}:file2]\n"
        "search = The current version is {current_version}\n"
        "replace = The current version is {new_version}\n",
        encoding="utf-8",
    )
    setup_cfg = config.get_configuration(cfg_file_path)
    assert len(setup_cfg.files) == 2
    assert all(f.filename == "file2" for f in setup_cfg.files)


def test_utf8_message_from_config_file(tmp_path: Path, cfg_file):
    cfg_path = tmp_path / cfg_file
    initial_config = (
        "[bumpversion]\n"
        "current_version = 500.0.0\n"
        "commit = True\n"
        "message = Nová verze: {current_version} ☃, {new_version} ☀\n"
    )
    cfg_path.write_bytes(initial_config.encode("utf-8"))

    with inside_dir(tmp_path):
        cfg = config.get_configuration(cfg_path)

    assert cfg.message == "Nová verze: {current_version} ☃, {new_version} ☀"


def test_current_version_is_always_a_string(tmp_path: Path):
    cfg_path = tmp_path / ".bumpversion.cfg"
    initial_config = (
        "[bumpversion]\n"
        "current_version = 2\n"
        "commit = False\n"
        "tag = False\n"
        "parse = (?P<major>\d+)\n"
        "serialize = \n"
        "	{major}\n"
    )
    cfg_path.write_bytes(initial_config.encode("utf-8"))

    with inside_dir(tmp_path):
        cfg = config.get_configuration(cfg_path)

    assert cfg.current_version == "2"



"""Tests for the serialization of versioned objects."""

from bumpversion.versioning.serialization import parse_version, serialize
from bumpversion.versioning.conventions import semver_spec, SEMVER_PATTERN
from bumpversion.versioning.models import Version, VersionSpec, VersionComponentSpec
from bumpversion.exceptions import BumpVersionError, FormattingError

import pytest
from pytest import param


class TestParseVersion:
    """Test the parse_version function."""

    def test_empty_version_returns_empty_dict(self):
        assert parse_version("", "") == {}
        assert parse_version(None, "") == {}

    def test_empty_parse_pattern_returns_empty_dict(self):
        assert parse_version("1.2.3", "") == {}
        assert parse_version("1.2.3", None) == {}

    @pytest.mark.parametrize(
        ["version_string", "parse_pattern", "expected"],
        [
            param(
                "1.2.3",
                SEMVER_PATTERN,
                {"buildmetadata": "", "major": "1", "minor": "2", "patch": "3", "pre_l": "", "pre_n": ""},
                id="parse-version",
            ),
            param(
                "1.2.3-alpha1",
                SEMVER_PATTERN,
                {"buildmetadata": "", "major": "1", "minor": "2", "patch": "3", "pre_l": "alpha", "pre_n": "1"},
                id="parse-prerelease",
            ),
            param(
                "1.2.3+build.1",
                SEMVER_PATTERN,
                {"buildmetadata": "build.1", "major": "1", "minor": "2", "patch": "3", "pre_l": "", "pre_n": ""},
                id="parse-buildmetadata",
            ),
            param(
                "1.2.3-alpha1+build.1",
                SEMVER_PATTERN,
                {"buildmetadata": "build.1", "major": "1", "minor": "2", "patch": "3", "pre_l": "alpha", "pre_n": "1"},
                id="parse-prerelease-buildmetadata",
            ),
        ],
    )
    def test_parses_version_and_returns_full_dict(self, version_string: str, parse_pattern: str, expected: dict):
        """The version string should be parsed into a dictionary of the parts and values, including missing parts."""
        results = parse_version(version_string, parse_pattern)
        assert results == expected

    def test_unmatched_pattern_returns_empty_dict(self):
        """If the version string doesn't match the parse pattern, an empty dictionary should be returned."""
        assert parse_version("1.2.3", r"v(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)") == {}

    def test_invalid_parse_pattern_raises_error(self):
        """If the parse pattern is not a valid regular expression, a ValueError should be raised."""
        with pytest.raises(BumpVersionError):
            parse_version("1.2.3", r"v(?P<major>\d+\.(?P<minor>\d+)\.(?P<patch>\d+)")

    def test_parse_pattern_with_newlines(self):
        """A parse pattern with newlines should be parsed correctly."""
        pattern = r"MAJOR=(?P<major>\d+)\nMINOR=(?P<minor>\d+)\nPATCH=(?P<patch>\d+)\n"
        assert parse_version("MAJOR=31\nMINOR=0\nPATCH=3\n", pattern) == {"major": "31", "minor": "0", "patch": "3"}

    def test_parse_pattern_with_zero_padding(self):
        """A parse pattern with zero padding should be parsed correctly."""
        pattern = r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)"
        assert parse_version("5.06.0", pattern) == {"major": "5", "minor": "06", "patch": "0"}


class TestSerialize:
    """Test the serialize function."""

    class TestFormatSelection:
        @pytest.mark.parametrize(
            ["version", "expected"],
            [
                param(
                    semver_spec().create_version({"major": "1", "minor": "2", "patch": "3"}),
                    "1.2.3",
                    id="major-minor-patch",
                ),
                param(
                    semver_spec().create_version({"major": "1", "minor": "2", "patch": "0"}),
                    "1.2",
                    id="major-minor-patch-zero",
                ),
                param(
                    semver_spec().create_version({"major": "1", "minor": "0", "patch": "0"}),
                    "1",
                    id="major-minor-zero-patch-zero",
                ),
            ],
        )
        def test_is_string_with_fewest_required_labels(self, version: Version, expected: str):
            """The smallest pattern with all the required values should be picked."""
            patterns = ["{major}.{minor}.{patch}", "{major}.{minor}", "{major}"]
            assert serialize(version, serialize_patterns=patterns, context={}) == expected

        def test_is_renderable_pattern_with_fewest_labels(self):
            """The smallest renderable pattern should be picked if no pattern has all the required values."""
            version = semver_spec().create_version(
                {"major": "1", "minor": "2", "patch": "3", "pre_l": "a", "buildmetadata": "build.1"}
            )
            # None of the patterns have all the required values, so the smallest renderable pattern should be picked
            assert (
                serialize(
                    version,
                    serialize_patterns=["{major}.{minor}.{patch}+{buildmetadata}", "{major}.{minor}.{patch}"],
                    context={},
                )
                == "1.2.3"
            )

        def test_is_first_pattern_with_fewest_labels(self):
            """The first pattern with the fewest labels should be picked if multiple have the same number of labels."""
            version = semver_spec().create_version(
                {"major": "1", "minor": "2", "patch": "3", "buildmetadata": "build.1"}
            )
            assert (
                serialize(
                    version,
                    serialize_patterns=["{major}.{minor}.{buildmetadata}", "{major}.{minor}.{patch}"],
                    context={},
                )
                == "1.2.build.1"
            )

        def test_includes_optional_component_when_dependent_is_not_optional(self):
            """A format with an optional component should render when the dependent component is not optional."""
            version_spec = VersionSpec(
                components={
                    "major": VersionComponentSpec(),
                    "minor": VersionComponentSpec(),
                    "patch": VersionComponentSpec(),
                    "release": VersionComponentSpec(
                        optional_value="g",
                        first_value="g",
                        values=[
                            "dev",
                            "a",
                            "b",
                            "g",
                        ],
                    ),
                    "build": VersionComponentSpec(),
                },
            )
            serialize_patterns = [
                "{major}.{minor}.{patch}{release}{build}",
                "{major}.{minor}.{patch}{release}",
                "{major}.{minor}.{patch}",
            ]

            version = version_spec.create_version({"major": "0", "minor": "3", "patch": "1", "build": "1"})
            assert (
                serialize(
                    version,
                    serialize_patterns=serialize_patterns,
                    context={},
                )
                == "0.3.1g1"
            )

        def test_selects_first_pattern_when_all_are_invalid(self):
            """If all patterns are invalid, the first pattern should be selected."""
            version = semver_spec().create_version({"major": "1", "minor": "2", "patch": "3"})
            assert (
                serialize(
                    version,
                    serialize_patterns=["{major}", "{major}.{minor}"],
                    context={},
                )
                == "1"
            )

    class TestRendersFormat:
        def test_with_newlines(self):
            """A serialization format with newlines should be rendered correctly."""
            version = semver_spec().create_version({"major": "31", "minor": "0", "patch": "3"})
            assert (
                serialize(version, serialize_patterns=["MAJOR={major}\nMINOR={minor}\nPATCH={patch}\n"], context={})
                == "MAJOR=31\nMINOR=0\nPATCH=3\n"
            )

        def test_with_additional_context(self):
            """A serialization format with additional context should be rendered correctly."""
            version = semver_spec().create_version({"major": "1", "minor": "2", "patch": "3"})
            assert (
                serialize(
                    version,
                    serialize_patterns=["{major}.{minor}.{patch}+{$BUILDMETADATA}"],
                    context={"$BUILDMETADATA": "build.1"},
                )
                == "1.2.3+build.1"
            )

        def test_zero_padding(self):
            """A numeric component with zero-padding notation should be rendered correctly."""
            version = semver_spec().create_version({"major": "1", "minor": "2", "patch": "3"})
            assert serialize(version, serialize_patterns=["{major}.{minor:02}.{patch}"], context={}) == "1.02.3"

    class TestRaisesError:
        def test_if_context_is_missing_values(self):
            """An error is raised if not all parts required in the format have values."""
            version = semver_spec().create_version({"major": "1", "minor": "2"})
            with pytest.raises(FormattingError):
                serialize(
                    version,
                    serialize_patterns=["{major}.{minor}.{patch}+{$BUILDMETADATA}"],
                    context={},
                )



"""Functions for displaying information about the version."""

import dataclasses
from io import StringIO
from pprint import pprint
from typing import Any, Optional

from bumpversion.bump import get_next_version
from bumpversion.config import Config
from bumpversion.context import get_context
from bumpversion.exceptions import BadInputError
from bumpversion.ui import print_error, print_info
from bumpversion.utils import recursive_sort_dict


def output_default(value: dict) -> None:
    """Output the value with key=value or just value if there is only one item."""
    if len(value) == 1:
        print_info(next(iter(value.values())))
    else:
        buffer = StringIO()
        pprint(value, stream=buffer, sort_dicts=True)  # noqa: T203
        print_info(buffer.getvalue())


def output_yaml(value: dict) -> None:
    """Output the value as yaml."""
    from bumpversion.yaml_dump import dump

    print_info(dump(recursive_sort_dict(value)))


def output_json(value: dict) -> None:
    """Output the value as json."""
    import json

    def default_encoder(obj: Any) -> str:
        if dataclasses.is_dataclass(obj):
            return str(obj)
        elif isinstance(obj, type):
            return obj.__name__
        raise TypeError(f"Object of type {type(obj), str(obj)} is not JSON serializable")

    print_info(json.dumps(value, sort_keys=True, indent=2, default=default_encoder))


OUTPUTTERS = {
    "yaml": output_yaml,
    "json": output_json,
    "default": output_default,
}


def resolve_name(obj: Any, name: str, default: Any = None, err_on_missing: bool = False) -> Any:
    """
    Get a key or attr ``name`` from obj or default value.

    Copied and modified from Django Template variable resolutions

    Resolution methods:

    - Mapping key lookup
    - Attribute lookup
    - Sequence index

    Args:
        obj: The object to access
        name: A dotted name to the value, such as ``mykey.0.name``
        default: If the name cannot be resolved from the object, return this value
        err_on_missing: Raise a `BadInputError` if the name cannot be resolved

    Returns:
        The value at the resolved name or the default value.

    Raises:
        BadInputError: If we cannot resolve the name and `err_on_missing` is `True`

    # noqa: DAR401
    """
    lookups = name.split(".")
    current = obj
    try:  # catch-all for unexpected failures
        for bit in lookups:
            try:  # dictionary lookup
                current = current[bit]
                # ValueError/IndexError are for numpy.array lookup on
                # numpy < 1.9 and 1.9+ respectively
            except (TypeError, AttributeError, KeyError, ValueError, IndexError):
                try:  # attribute lookup
                    current = getattr(current, bit)
                except (TypeError, AttributeError):
                    # Reraise if the exception was raised by a @property
                    if bit in dir(current):
                        raise
                    try:  # list-index lookup
                        current = current[int(bit)]
                    except (
                        IndexError,  # list index out of range
                        ValueError,  # invalid literal for int()
                        KeyError,  # current is a dict without `int(bit)` key
                        TypeError,
                    ):  # un-subscript-able object
                        return default
        return current
    except Exception as e:  # noqa: BLE001 # pragma: no cover
        if err_on_missing:
            raise BadInputError(f"Could not resolve '{name}'") from e
        else:
            return default


def do_show(*args, config: Config, format_: str = "default", increment: Optional[str] = None) -> None:
    """Show current version or configuration information."""
    config_dict = config.model_dump()
    ctx = get_context(config)

    if increment:
        version = config.version_config.parse(config.current_version)
        next_version = get_next_version(version, config, increment, None)
        next_version_str = config.version_config.serialize(next_version, ctx)
        config_dict["new_version"] = next_version_str

    try:
        if "all" in args or not args:
            show_items = config_dict
        else:
            show_items = {key: resolve_name(config_dict, key) for key in args}

        OUTPUTTERS.get(format_, OUTPUTTERS["default"])(show_items)
    except BadInputError as e:
        print_error(e.message)



import shutil
from pathlib import Path

from click.testing import Result

from bumpversion import cli
from tests.conftest import inside_dir


def test_show(tmp_path: Path, fixtures_path: Path, runner):
    """The show subcommand should list the parts of the configuration."""
    # Arrange
    config_path = tmp_path / "pyproject.toml"
    toml_path = fixtures_path / "basic_cfg.toml"
    shutil.copy(toml_path, config_path)

    with inside_dir(tmp_path):
        result: Result = runner.invoke(cli.cli, ["show", "current_version"])

    if result.exit_code != 0:
        print(result.output)
        print(result.exception)

    assert result.exit_code == 0
    assert result.output.splitlines(keepends=False) == ["1.0.0"]


def test_show_and_increment(tmp_path: Path, fixtures_path: Path, runner):
    """The show subcommand should increment the version and display it."""
    # Arrange
    config_path = tmp_path / "pyproject.toml"
    toml_path = fixtures_path / "basic_cfg.toml"
    shutil.copy(toml_path, config_path)

    with inside_dir(tmp_path):
        result: Result = runner.invoke(cli.cli, ["show", "new_version", "--increment", "minor"])

    if result.exit_code != 0:
        print(result.output)
        print(result.exception)

    assert result.exit_code == 0
    assert result.output.splitlines(keepends=False) == ["1.1.0-dev"]


def test_show_no_args(tmp_path: Path, fixtures_path: Path, runner):
    """The show subcommand should list the entire configuration."""
    # Arrange
    config_path = tmp_path / "pyproject.toml"
    toml_path = fixtures_path / "basic_cfg.toml"
    expected_output = fixtures_path.joinpath("basic_cfg_expected.yaml").read_text()

    shutil.copy(toml_path, config_path)

    with inside_dir(tmp_path):
        result: Result = runner.invoke(cli.cli, ["show", "--format", "yaml"])

    if result.exit_code != 0:
        print(result.output)
        print(result.exception)

    assert result.exit_code == 0
    assert result.output.strip() == expected_output.strip()



"""Functions for serializing and deserializing version objects."""

import re
from copy import copy
from operator import itemgetter
from typing import Dict, List, MutableMapping

from bumpversion.exceptions import BumpVersionError, FormattingError
from bumpversion.ui import get_indented_logger
from bumpversion.utils import key_val_string, labels_for_format
from bumpversion.versioning.models import Version

logger = get_indented_logger(__name__)


def parse_version(version_string: str, parse_pattern: str) -> Dict[str, str]:
    """
    Parse a version string into a dictionary of the parts and values using a regular expression.

    Args:
        version_string: Version string to parse
        parse_pattern: The regular expression pattern to use for parsing

    Returns:
        A dictionary of version part labels and their values, or an empty dictionary
        if the version string doesn't match.

    Raises:
        BumpVersionError: If the parse_pattern is not a valid regular expression
    """
    if not version_string:
        logger.debug("Version string is empty, returning empty dict")
        return {}
    elif not parse_pattern:
        logger.debug("Parse pattern is empty, returning empty dict")
        return {}

    logger.debug("Parsing version '%s' using regexp '%s'", version_string, parse_pattern)
    logger.indent()

    try:
        pattern = re.compile(parse_pattern, re.VERBOSE)
    except re.error as e:
        raise BumpVersionError(f"'{parse_pattern}' is not a valid regular expression.") from e

    match = re.search(pattern, version_string)

    if not match:
        logger.debug(
            "'%s' does not parse current version '%s'",
            parse_pattern,
            version_string,
        )
        return {}

    parsed = match.groupdict(default="")
    logger.debug("Parsed the following values: %s", key_val_string(parsed))
    logger.dedent()

    return parsed


def multisort(xs: list, specs: tuple) -> list:
    """
    Sort a list of dictionaries by multiple keys.

    From https://docs.python.org/3/howto/sorting.html#sort-stability-and-complex-sorts

    Args:
        xs: The list of dictionaries to sort
        specs: A tuple of (key, reverse) pairs

    Returns:
        The sorted list
    """
    for key, reverse in reversed(specs):
        xs.sort(key=itemgetter(key), reverse=reverse)
    return xs


def serialize(version: Version, serialize_patterns: List[str], context: MutableMapping) -> str:
    """
    Attempts to serialize a version with the given serialization format.

    - valid serialization patterns are those that are renderable with the given context
    - formats that contain all required components are preferred
    - the shortest valid serialization pattern is used
    - if two patterns are equally short, the first one is used
    - if no valid serialization pattern is found, an error is raised

    Args:
        version: The version to serialize
        serialize_patterns: The serialization format to use, using Python's format string syntax
        context: The context to use when serializing the version

    Raises:
        FormattingError: if a serialization pattern

    Returns:
        The serialized version as a string
    """
    logger.debug("Serializing version '%s'", version)
    logger.indent()

    local_context = copy(context)
    local_context.update(version.values())
    local_context_keys = set(local_context.keys())
    required_component_labels = set(version.required_components())

    patterns = []
    for index, pattern in enumerate(serialize_patterns):
        labels = set(labels_for_format(pattern))
        patterns.append(
            {
                "pattern": pattern,
                "labels": labels,
                "order": index,
                "num_labels": len(labels),
                "renderable": local_context_keys >= labels,
                "has_required_components": required_component_labels <= labels,
            }
        )

    valid_patterns = filter(itemgetter("renderable"), patterns)
    sorted_patterns = multisort(
        list(valid_patterns), (("has_required_components", True), ("num_labels", False), ("order", False))
    )

    if not sorted_patterns:
        raise FormattingError(f"Could not find a valid serialization format in {serialize_patterns!r} for {version!r}")

    chosen_pattern = sorted_patterns[0]["pattern"]
    logger.debug("Using serialization format '%s'", chosen_pattern)
    serialized = chosen_pattern.format(**local_context)
    logger.debug("Serialized to '%s'", serialized)
    logger.dedent()

    return serialized



"""Module for managing Versions and their internal parts."""

import re
from typing import Any, Dict, List, MutableMapping, Optional, Tuple

from click import UsageError

from bumpversion.ui import get_indented_logger
from bumpversion.utils import labels_for_format
from bumpversion.versioning.models import Version, VersionComponentSpec, VersionSpec
from bumpversion.versioning.serialization import parse_version, serialize

logger = get_indented_logger(__name__)


class VersionConfig:
    """
    Hold a complete representation of a version string.
    """

    def __init__(
        self,
        parse: str,
        serialize: Tuple[str],
        search: str,
        replace: str,
        part_configs: Optional[Dict[str, VersionComponentSpec]] = None,
    ):
        try:
            self.parse_regex = re.compile(parse, re.VERBOSE)
        except re.error as e:
            raise UsageError(f"--parse '{parse}' is not a valid regex.") from e

        self.serialize_formats = serialize
        self.part_configs = part_configs or {}
        self.version_spec = VersionSpec(self.part_configs)
        # TODO: I think these two should be removed from the config object
        self.search = search
        self.replace = replace

    def __repr__(self) -> str:
        return f"<bumpversion.VersionConfig:{self.parse_regex.pattern}:{self.serialize_formats}>"

    def __eq__(self, other: Any) -> bool:
        return (
            self.parse_regex.pattern == other.parse_regex.pattern
            and self.serialize_formats == other.serialize_formats
            and self.part_configs == other.part_configs
            and self.search == other.search
            and self.replace == other.replace
        )

    @property
    def order(self) -> List[str]:
        """
        Return the order of the labels in a serialization format.

        Currently, order depends on the first given serialization format.
        This seems like a good idea because this should be the most complete format.

        Returns:
            A list of version part labels in the order they should be rendered.
        """
        return labels_for_format(self.serialize_formats[0])

    def parse(self, version_string: Optional[str] = None) -> Optional[Version]:
        """
        Parse a version string into a Version object.

        Args:
            version_string: Version string to parse

        Returns:
            A Version object representing the string.
        """
        parsed = parse_version(version_string, self.parse_regex.pattern)

        if not parsed:
            return None

        version = self.version_spec.create_version(parsed)
        version.original = version_string
        return version

    def serialize(self, version: Version, context: MutableMapping) -> str:
        """
        Serialize a version to a string.

        Args:
            version: The version to serialize
            context: The context to use when serializing the version

        Returns:
            The serialized version as a string
        """
        return serialize(version, list(self.serialize_formats), context)



"""Tests for the Version class."""

from freezegun import freeze_time

from bumpversion.versioning.models import VersionComponentSpec
from bumpversion.versioning.models import VersionSpec
import pytest
from pytest import param


@pytest.fixture
def semver_version_spec():
    """Return a version spec."""
    config = {
        "major": VersionComponentSpec(),
        "minor": VersionComponentSpec(),
        "patch": VersionComponentSpec(),
        "build": VersionComponentSpec(optional_value="0", independent=True),
        "auto": VersionComponentSpec(optional_value="0", independent=True, always_increment=True),
    }

    return VersionSpec(config)


@pytest.fixture
def calver_version_spec():
    """Return a calver version spec."""
    config = {
        "release": VersionComponentSpec(calver_format="{YYYY}.{MM}.{DD}"),
        "patch": VersionComponentSpec(),
        "build": VersionComponentSpec(optional_value="0", independent=True),
    }

    return VersionSpec(config)


class TestSemVerVersion:
    """Test how a SemVer-style Version model should behave."""

    def test_acts_like_a_dict(self, semver_version_spec: VersionSpec):
        """You can get version components by name."""
        # Arrange
        version = semver_version_spec.create_version({"major": "1", "minor": "2", "patch": "3"})

        # Assert
        assert version["major"].value == "1"
        assert version["minor"].value == "2"
        assert version["patch"].value == "3"
        assert version["build"].value == "0"
        assert version["auto"].value == "0"

        with pytest.raises(KeyError):
            version["invalid"]

    def test_length_is_number_of_parts(self, semver_version_spec: VersionSpec):
        # Arrange
        version = semver_version_spec.create_version({"major": "1"})

        # Assert
        assert len(version) == 5

    def test_is_iterable(self, semver_version_spec: VersionSpec):
        # Arrange
        version = semver_version_spec.create_version({"major": "1"})

        # Assert
        assert list(version) == ["major", "minor", "patch", "build", "auto"]

    def test_has_a_string_representation(self, semver_version_spec: VersionSpec):
        # Arrange
        version = semver_version_spec.create_version({"major": "1", "minor": "2", "patch": "3"})

        # Assert
        assert repr(version) == "<bumpversion.Version:auto=0, build=0, major=1, minor=2, patch=3>"

    def test_has_an_equality_comparison(self, semver_version_spec: VersionSpec):
        # Arrange
        version1 = semver_version_spec.create_version({"major": "1", "minor": "2", "patch": "3"})
        version2 = semver_version_spec.create_version({"major": "1", "minor": "2", "patch": "3"})
        version3 = semver_version_spec.create_version({"major": "1", "minor": "2", "patch": "4"})

        # Assert
        assert version1 == version2
        assert version1 != version3

    class TestBump:
        """Tests of the bump method."""

        def test_does_not_change_original_version(self, semver_version_spec: VersionSpec):
            """Bumping a version does not change the original version."""
            # Arrange
            version1 = semver_version_spec.create_version({"major": "1", "minor": "2", "patch": "3"})

            # Act
            version2 = version1.bump("patch")

            # Assert
            assert version1["patch"].value == "3"
            assert version2["patch"].value == "4"

        def test_returns_a_new_version(self, semver_version_spec: VersionSpec):
            """Bumping a version returns a new version."""
            # Arrange
            version1 = semver_version_spec.create_version({"major": "1", "minor": "2", "patch": "3"})

            # Act
            version2 = version1.bump("patch")

            # Assert
            assert version1 is not version2

        def test_changes_component_and_dependents(self, semver_version_spec: VersionSpec):
            """Bumping a version bumps the specified component and changes its dependents."""
            # Arrange
            version1 = semver_version_spec.create_version(
                {"major": "1", "minor": "2", "patch": "3", "build": "4", "auto": "5"}
            )

            # Act
            patch_version = version1.bump("patch")
            minor_version = version1.bump("minor")
            major_version = version1.bump("major")
            build_version = version1.bump("build")

            # Assert
            patch_version_str = ".".join([item.value for item in patch_version.components.values()])
            minor_version_str = ".".join([item.value for item in minor_version.components.values()])
            major_version_str = ".".join([item.value for item in major_version.components.values()])
            build_version_str = ".".join([item.value for item in build_version.components.values()])
            assert patch_version_str == "1.2.4.4.6"
            assert minor_version_str == "1.3.0.4.6"
            assert major_version_str == "2.0.0.4.6"
            assert build_version_str == "1.2.3.5.6"

    class TestRequiredComponents:
        """Tests of the required_keys function."""

        @pytest.mark.parametrize(
            ["values", "expected"],
            [
                param({"major": "1", "minor": "2", "patch": "3"}, ["major", "minor", "patch"], id="major-minor-patch"),
                param(
                    {"major": "0", "minor": "2", "patch": "3", "build": "4"},
                    ["minor", "patch", "build"],
                    id="minor-patch-build",
                ),
                param({"major": "1", "minor": "0", "patch": "3", "build": ""}, ["major", "patch"], id="major-patch"),
                param({"major": "0", "minor": "0", "patch": "3", "build": ""}, ["patch"], id="patch"),
            ],
        )
        def test_returns_required_component_names(
            self, semver_version_spec: VersionSpec, values: dict, expected: list
        ):
            """The required_keys function returns all required keys."""
            # Arrange
            version = semver_version_spec.create_version(values)

            # Act
            required_components = version.required_components()

            # Assert
            assert required_components == expected


class TestCalVerVersion:
    """Test how a CalVer-style Version model should behave."""

    @freeze_time("2020-05-01")
    def test_acts_like_a_dict(self, calver_version_spec: VersionSpec):
        """You can get version components by name."""
        # Arrange
        version = calver_version_spec.create_version({"release": "2020.4.1", "patch": "3", "build": "10"})

        # Assert
        assert version["release"].value == "2020.4.1"
        assert version["patch"].value == "3"
        assert version["build"].value == "10"

        with pytest.raises(KeyError):
            version["invalid"]

    def test_length_is_number_of_parts(self, calver_version_spec: VersionSpec):
        # Arrange
        version = calver_version_spec.create_version({"release": "2020.4.1"})

        # Assert
        assert len(version) == 3

    def test_version_is_iterable(self, calver_version_spec: VersionSpec):
        # Arrange
        version = calver_version_spec.create_version({"release": "2020.4.1"})

        # Assert
        assert list(version) == ["release", "patch", "build"]

    def test_has_a_string_representation(self, calver_version_spec: VersionSpec):
        # Arrange
        version = calver_version_spec.create_version({"release": "2020.4.1", "patch": "3"})

        # Assert
        assert repr(version) == "<bumpversion.Version:build=0, patch=3, release=2020.4.1>"

    def test_version_has_an_equality_comparison(self, calver_version_spec: VersionSpec):
        # Arrange
        version1 = calver_version_spec.create_version({"release": "2020.4.1", "patch": "3"})
        version2 = calver_version_spec.create_version({"release": "2020.4.1", "patch": "3"})
        version3 = calver_version_spec.create_version({"release": "2020.5.1"})

        # Assert
        assert version1 == version2
        assert version1 != version3

    class TestBump:
        """Tests of the bump method."""

        @freeze_time("2020-05-01")
        def test_bump_does_not_change_original_version(self, calver_version_spec: VersionSpec):
            """Bumping a version does not change the original version."""
            # Arrange
            version1 = calver_version_spec.create_version({"release": "2020.4.1", "patch": "3", "build": "10"})

            # Act
            version2 = version1.bump("patch")

            # Assert
            assert version1["patch"].value == "3"
            assert version1["release"].value == "2020.4.1"
            assert version2["patch"].value == "0"
            assert version2["release"].value == "2020.5.1"

        @freeze_time("2020-05-01")
        def test_bump_returns_a_new_version(self, calver_version_spec: VersionSpec):
            """Bumping a version returns a new version."""
            # Arrange
            version1 = calver_version_spec.create_version({"release": "2020.4.1", "patch": "3"})

            # Act
            version2 = version1.bump("patch")

            assert version2["release"].value == "2020.5.1"
            assert version2["patch"].value == "0"

            # Assert
            assert version1 is not version2

        @freeze_time("2020-05-01")
        def test_bump_always_increments_calver(self, calver_version_spec: VersionSpec):
            """Bumping a version always increments the calver."""
            # Arrange
            version1 = calver_version_spec.create_version({"release": "2020.4.1", "patch": "3", "build": "10"})

            # Act
            version2 = version1.bump("patch")

            # Assert
            assert version2["release"].value == "2020.5.1"
            assert version2["patch"].value == "0"

        @freeze_time("2020-05-01")
        def test_bump_changes_component_and_dependents(self, calver_version_spec: VersionSpec):
            """Bumping a version bumps the specified component and changes its dependents."""
            # Arrange
            version1 = calver_version_spec.create_version({"release": "2020.5.1", "patch": "3", "build": "4"})

            # Act
            patch_version = version1.bump("patch")
            release_version = version1.bump("release")
            build_version = version1.bump("build")

            # Assert
            patch_version_str = ".".join([item.value for item in patch_version.components.values()])
            release_version_str = ".".join([item.value for item in release_version.components.values()])
            build_version_str = ".".join([item.value for item in build_version.components.values()])
            assert patch_version_str == "2020.5.1.4.4"
            assert release_version_str == "2020.5.1.0.4"
            assert build_version_str == "2020.5.1.3.5"

    class TestRequiredComponents:
        """Tests of the required_keys function."""

        @pytest.mark.parametrize(
            ["values", "expected"],
            [
                param({"release": "2020.4.1", "patch": "3"}, ["release", "patch"], id="release-patch"),
                param({"release": "2020.4.1", "build": "4"}, ["release", "build"], id="release-build"),
                param({"release": "2020.4.1"}, ["release"], id="release"),
            ],
        )
        @freeze_time("2020-05-01")
        def test_returns_required_component_names(
            self, calver_version_spec: VersionSpec, values: dict, expected: list
        ):
            """The required_keys function returns all required keys."""
            # Arrange
            version = calver_version_spec.create_version(values)

            # Act
            required_components = version.required_components()

            # Assert
            assert required_components == expected



"""Bump My Version configuration models."""

from __future__ import annotations

import re
from collections import defaultdict
from itertools import chain
from typing import TYPE_CHECKING, Dict, List, MutableMapping, Optional, Tuple, Union

from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from bumpversion.ui import get_indented_logger
from bumpversion.versioning.models import VersionComponentSpec  # NOQA: TCH001

if TYPE_CHECKING:
    from bumpversion.scm import SCMInfo
    from bumpversion.version_part import VersionConfig
    from bumpversion.versioning.models import VersionSpec

logger = get_indented_logger(__name__)


class FileChange(BaseModel):
    """A change to make to a file."""

    parse: str
    serialize: tuple
    search: str
    replace: str
    regex: bool
    ignore_missing_version: bool
    ignore_missing_file: bool
    filename: Optional[str] = None
    glob: Optional[str] = None  # Conflicts with filename. If both are specified, glob wins
    key_path: Optional[str] = None  # If specified, and has an appropriate extension, will be treated as a data file

    def __hash__(self):
        """Return a hash of the model."""
        return hash(tuple(sorted(self.model_dump().items())))

    def get_search_pattern(self, context: MutableMapping) -> Tuple[re.Pattern, str]:
        """
        Render the search pattern and return the compiled regex pattern and the raw pattern.

        Args:
            context: The context to use for rendering the search pattern

        Returns:
            A tuple of the compiled regex pattern and the raw pattern as a string.
        """
        logger.debug("Rendering search pattern with context")
        logger.indent()
        # the default search pattern is escaped, so we can still use it in a regex
        raw_pattern = self.search.format(**context)
        default = re.compile(re.escape(raw_pattern), re.MULTILINE | re.DOTALL)
        if not self.regex:
            logger.debug("No RegEx flag detected. Searching for the default pattern: '%s'", default.pattern)
            logger.dedent()
            return default, raw_pattern

        re_context = {key: re.escape(str(value)) for key, value in context.items()}
        regex_pattern = self.search.format(**re_context)
        try:
            search_for_re = re.compile(regex_pattern, re.MULTILINE | re.DOTALL)
            logger.debug("Searching for the regex: '%s'", search_for_re.pattern)
            logger.dedent()
            return search_for_re, raw_pattern
        except re.error as e:
            logger.error("Invalid regex '%s': %s.", default, e)

        logger.debug("Invalid regex. Searching for the default pattern: '%s'", raw_pattern)
        logger.dedent()

        return default, raw_pattern


class Config(BaseSettings):
    """Bump Version configuration."""

    current_version: Optional[str]
    parse: str
    serialize: tuple = Field(min_length=1)
    search: str
    replace: str
    regex: bool
    ignore_missing_version: bool
    ignore_missing_files: bool
    tag: bool
    sign_tags: bool
    tag_name: str
    tag_message: Optional[str]
    allow_dirty: bool
    commit: bool
    message: str
    commit_args: Optional[str]
    scm_info: Optional["SCMInfo"]
    parts: Dict[str, VersionComponentSpec]
    files: List[FileChange] = Field(default_factory=list)
    included_paths: List[str] = Field(default_factory=list)
    excluded_paths: List[str] = Field(default_factory=list)
    model_config = SettingsConfigDict(env_prefix="bumpversion_")
    _resolved_filemap: Optional[Dict[str, List[FileChange]]] = None

    def add_files(self, filename: Union[str, List[str]]) -> None:
        """Add a filename to the list of files."""
        filenames = [filename] if isinstance(filename, str) else filename
        files = set(self.files)
        for name in filenames:
            files.add(
                FileChange(
                    filename=name,
                    glob=None,
                    key_path=None,
                    parse=self.parse,
                    serialize=self.serialize,
                    search=self.search,
                    replace=self.replace,
                    regex=self.regex,
                    ignore_missing_version=self.ignore_missing_version,
                    ignore_missing_file=self.ignore_missing_files,
                )
            )
        self.files = list(files)

        self._resolved_filemap = None

    @property
    def resolved_filemap(self) -> Dict[str, List[FileChange]]:
        """Return the cached resolved filemap."""
        if self._resolved_filemap is None:
            self._resolved_filemap = self._resolve_filemap()
        return self._resolved_filemap

    def _resolve_filemap(self) -> Dict[str, List[FileChange]]:
        """Return a map of filenames to file configs, expanding any globs."""
        from bumpversion.config.utils import resolve_glob_files

        output = defaultdict(list)
        new_files = []
        for file_cfg in self.files:
            if file_cfg.glob:
                new_files.extend(resolve_glob_files(file_cfg))
            else:
                new_files.append(file_cfg)

        for file_cfg in new_files:
            output[file_cfg.filename].append(file_cfg)
        return output

    @property
    def files_to_modify(self) -> List[FileChange]:
        """Return a list of files to modify."""
        files_not_excluded = [filename for filename in self.resolved_filemap if filename not in self.excluded_paths]
        inclusion_set = set(self.included_paths) | set(files_not_excluded)
        return list(
            chain.from_iterable(
                file_cfg_list for key, file_cfg_list in self.resolved_filemap.items() if key in inclusion_set
            )
        )

    @property
    def version_config(self) -> "VersionConfig":
        """Return the version configuration."""
        from bumpversion.version_part import VersionConfig

        return VersionConfig(self.parse, self.serialize, self.search, self.replace, self.parts)

    def version_spec(self, version: Optional[str] = None) -> "VersionSpec":
        """Return the version specification."""
        from bumpversion.versioning.models import VersionSpec

        return VersionSpec(self.parts)



"""Utilities for user interface."""

import logging

import click
from click import UsageError, secho
from rich.logging import RichHandler

from bumpversion.indented_logger import IndentedLoggerAdapter

logger = logging.getLogger("bumpversion")

VERBOSITY = {
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.DEBUG,
}


def get_indented_logger(name: str) -> "IndentedLoggerAdapter":
    """Get a logger with indentation."""
    return IndentedLoggerAdapter(logging.getLogger(name))


def setup_logging(verbose: int = 0) -> None:
    """Configure the logging."""
    logging.basicConfig(
        level=VERBOSITY.get(verbose, logging.DEBUG),
        format="%(message)s",
        datefmt="[%X]",
        handlers=[
            RichHandler(
                rich_tracebacks=True, show_level=False, show_path=False, show_time=False, tracebacks_suppress=[click]
            )
        ],
    )
    root_logger = get_indented_logger("")
    root_logger.setLevel(VERBOSITY.get(verbose, logging.DEBUG))


def print_info(msg: str) -> None:
    """Echo a message to the console."""
    secho(msg)


def print_error(msg: str) -> None:
    """Raise an error and exit."""
    raise UsageError(msg)


def print_warning(msg: str) -> None:
    """Echo a warning to the console."""
    secho(f"\nWARNING:\n\n{msg}\n", fg="yellow")



"""Configuration management."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union

from bumpversion.config.files import read_config_file
from bumpversion.config.models import Config
from bumpversion.exceptions import ConfigurationError
from bumpversion.ui import get_indented_logger

if TYPE_CHECKING:  # pragma: no-coverage
    from pathlib import Path

logger = get_indented_logger(__name__)

DEFAULTS = {
    "current_version": None,
    "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)",
    "serialize": ("{major}.{minor}.{patch}",),
    "search": "{current_version}",
    "replace": "{new_version}",
    "regex": False,
    "ignore_missing_version": False,
    "ignore_missing_files": False,
    "tag": False,
    "sign_tags": False,
    "tag_name": "v{new_version}",
    "tag_message": "Bump version: {current_version} → {new_version}",
    "allow_dirty": False,
    "commit": False,
    "message": "Bump version: {current_version} → {new_version}",
    "commit_args": None,
    "scm_info": None,
    "parts": {},
    "files": [],
}


def get_configuration(config_file: Union[str, Path, None] = None, **overrides: Any) -> Config:
    """
    Return the configuration based on any configuration files and overrides.

    Args:
        config_file: An explicit configuration file to use, otherwise search for one
        **overrides: Specific configuration key-values to override in the configuration

    Returns:
        The configuration
    """
    from bumpversion.config.utils import get_all_file_configs, get_all_part_configs
    from bumpversion.scm import SCMInfo, SourceCodeManager, get_scm_info  # noqa: F401

    logger.info("Reading configuration")
    logger.indent()

    config_dict = DEFAULTS.copy()
    parsed_config = read_config_file(config_file) if config_file else {}

    # We want to strip out unrecognized key-values to avoid inadvertent issues
    config_dict.update({key: val for key, val in parsed_config.items() if key in DEFAULTS.keys()})

    allowed_overrides = set(DEFAULTS.keys())
    config_dict.update({key: val for key, val in overrides.items() if key in allowed_overrides})

    # Set any missing version parts
    config_dict["parts"] = get_all_part_configs(config_dict)

    # Set any missing file configuration
    config_dict["files"] = get_all_file_configs(config_dict)

    # Resolve the SCMInfo class for Pydantic's BaseSettings
    Config.model_rebuild()
    config = Config(**config_dict)  # type: ignore[arg-type]

    # Get the information about the SCM
    scm_info = get_scm_info(config.tag_name, config.parse)
    config.scm_info = scm_info

    # Update and verify the current_version
    config.current_version = check_current_version(config)

    logger.dedent()

    return config


def check_current_version(config: Config) -> str:
    """
    Returns the current version.

    If the current version is not specified in the config file, command line or env variable,
    it attempts to retrieve it via a tag.

    Args:
        config: The current configuration dictionary.

    Returns:
        The version number

    Raises:
        ConfigurationError: If it can't find the current version
    """
    current_version = config.current_version
    scm_info = config.scm_info

    if current_version is None and scm_info.current_version:
        return scm_info.current_version
    elif current_version and scm_info.current_version and current_version != scm_info.current_version:
        logger.warning(
            "Specified version (%s) does not match last tagged version (%s)",
            current_version,
            scm_info.current_version,
        )
        return current_version
    elif current_version:
        return current_version

    raise ConfigurationError("Unable to determine the current version.")



"""Tests for the base CLI commands."""

from bumpversion import cli


class TestBadOptionRaisesError:
    """Tests to ensure that bad options raise an error."""

    def test_bad_option_raises_error(self, runner):
        """Passing an invalid option should raise an error."""
        result = runner.invoke(cli.cli, ["--bad-option", "bump", "--current-version", "1.0.0", "patch"])
        assert result.exit_code != 0
        assert "No such option" in result.output



"""Tests the show_bump command."""

import shutil
from pathlib import Path

from click.testing import Result
from bumpversion import cli
from tests.conftest import inside_dir


def test_show_bump_uses_current_version(tmp_path: Path, fixtures_path: Path, runner):
    """The show_bump subcommand should list the parts of the configuration."""
    # Arrange
    config_path = tmp_path / "pyproject.toml"
    toml_path = fixtures_path / "basic_cfg.toml"
    shutil.copy(toml_path, config_path)

    with inside_dir(tmp_path):
        result: Result = runner.invoke(cli.cli, ["show-bump"])

    if result.exit_code != 0:
        print(result.output)
        print(result.exception)

    assert result.exit_code == 0
    assert result.output == "\n".join(
        [
            "1.0.0 ── bump ─┬─ major ─── 2.0.0-dev",
            "               ├─ minor ─── 1.1.0-dev",
            "               ├─ patch ─── 1.0.1-dev",
            "               ╰─ release ─ invalid: The part has already the maximum value among "
            "['dev', 'gamma'] and cannot be bumped.\n",
        ]
    )


def test_show_bump_uses_passed_version(tmp_path: Path, fixtures_path: Path, runner):
    """The show_bump subcommand should list the parts of the configuration."""
    # Arrange
    config_path = tmp_path / "pyproject.toml"
    toml_path = fixtures_path / "basic_cfg.toml"
    shutil.copy(toml_path, config_path)

    with inside_dir(tmp_path):
        result: Result = runner.invoke(cli.cli, ["show-bump", "1.2.3"])

    if result.exit_code != 0:
        print(result.output)
        print(result.exception)

    assert result.exit_code == 0
    assert result.output == "\n".join(
        [
            "1.2.3 ── bump ─┬─ major ─── 2.0.0-dev",
            "               ├─ minor ─── 1.3.0-dev",
            "               ├─ patch ─── 1.2.4-dev",
            "               ╰─ release ─ invalid: The part has already the maximum value among ['dev', 'gamma'] "
            "and cannot be bumped.\n",
        ]
    )
